# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Analyse handlers."""


import importlib
import importlib.machinery
import importlib.util
import types
from importlib.machinery import ModuleSpec
from pathlib import Path
from typing import List, cast

import yaml

from autonomy.analyse.constants import CLASS_NAME, HANDLERS, HANDLERS_FILE


def load_handler_module_from_skill_path(skill_path: Path) -> types.ModuleType:
    """Resolve handler path to current workind directory."""

    loader = importlib.machinery.SourceFileLoader(
        HANDLERS, str(skill_path / "handlers.py")
    )
    spec = cast(ModuleSpec, importlib.util.spec_from_loader(HANDLERS, loader))
    module = importlib.util.module_from_spec(spec)
    loader.exec_module(module)

    return module


def check_handlers(config_file: Path, common_handlers: List[str]) -> None:
    """Check handlers"""

    if not (config_file.parent / HANDLERS_FILE).exists():
        raise FileNotFoundError(f"Handler file does not exist in {config_file.parent}")

    module = load_handler_module_from_skill_path(config_file.parent)
    module_attributes = dir(module)

    with open(str(config_file), mode="r", encoding="utf-8") as fp:
        config = yaml.safe_load(fp)
        for common_handler in common_handlers:
            if common_handler not in config[HANDLERS]:
                raise ValueError(
                    f"Common handler '{common_handler}' is not defined in {config_file}"
                )

        for handler_info in config[HANDLERS].values():
            if handler_info["class_name"] not in module_attributes:
                raise ValueError(
                    f"Handler {handler_info[CLASS_NAME]} declared in {config_file} is missing from {config_file.parent / HANDLERS}.py"
                )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test analyse ABCI docstrings."""

import difflib
import inspect
from pathlib import Path

from autonomy.analyse.abci.docstrings import (
    compare_docstring_content,
    docstring_abci_app,
)

from packages.valory.skills.hello_world_abci.rounds import HelloWorldAbciApp


def test_docstring_abci_app() -> None:
    """Test docstring_abci_app"""

    expected = """\"\"\"HelloWorldAbciApp

        Initial round: RegistrationRound

        Initial states: {RegistrationRound}

        Transition states:
            0. RegistrationRound
                - done: 1.
            1. CollectRandomnessRound
                - done: 2.
                - no majority: 1.
                - round timeout: 1.
            2. SelectKeeperRound
                - done: 3.
                - no majority: 0.
                - round timeout: 0.
            3. PrintMessageRound
                - done: 4.
                - round timeout: 0.
            4. ResetAndPauseRound
                - done: 1.
                - no majority: 0.
                - reset timeout: 0.

        Final states: {}

        Timeouts:
            round timeout: 30.0
            reset timeout: 30.0
    \"\"\""""

    docstring = docstring_abci_app(HelloWorldAbciApp)
    differences = "\n".join(difflib.unified_diff(docstring.split(), expected.split()))
    assert not differences, differences


def test_compare_docstring_content() -> None:
    """Test compare_docstring_content"""

    # no regex match
    assert compare_docstring_content("", "", "") == (False, "")

    # identical - no update
    docstring = docstring_abci_app(HelloWorldAbciApp)
    abci_app_name = HelloWorldAbciApp.__name__
    file_content = Path(inspect.getfile(HelloWorldAbciApp)).read_text()
    result = compare_docstring_content(file_content, docstring, abci_app_name)
    assert result == (True, file_content)

    # mutated - update
    mutated_content = file_content.replace("Initial round: RegistrationRound", "")
    assert not mutated_content == file_content
    result = compare_docstring_content(mutated_content, docstring, abci_app_name)
    assert result == (True, file_content)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Conftest module for Pytest."""
import inspect
import json
import os
import platform
import subprocess  # nosec
from pathlib import Path
from typing import Optional

import pytest
from aea.configurations.constants import PACKAGES
from aea.package_manager.base import PACKAGES_FILE


# https://pytest-cov.readthedocs.io/en/latest/subprocess-support.html#if-you-use-multiprocessing-process
try:
    from pytest_cov.embed import cleanup_on_sigterm  # type: ignore
except ImportError:
    pass
else:
    cleanup_on_sigterm()


CUR_PATH = os.path.dirname(inspect.getfile(inspect.currentframe()))  # type: ignore
ROOT_DIR = Path(CUR_PATH, "..").resolve().absolute()
DATA_DIR = ROOT_DIR / "tests" / "data"

skip_docker_tests = pytest.mark.skipif(
    platform.system() != "Linux",
    reason="Docker daemon is not available in Windows and macOS CI containers.",
)


def get_latest_git_tag() -> str:
    """Get the latest git tag"""
    res = subprocess.run(  # nosec
        [
            "git",
            "tag",
            "--sort=-committerdate",
        ],  # sort by commit date in descending order
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        check=True,
    )

    stdout = res.stdout.decode("utf-8")
    latest, *_ = stdout.split("\n")
    return latest.strip()


def get_file_from_tag(file_path: str, latest_tag: Optional[str] = None) -> str:
    """Get a specific file version from the commit history given a tag/commit"""
    latest_tag = latest_tag or get_latest_git_tag()
    res = subprocess.run(  # nosec
        ["git", "show", f"{latest_tag}:{file_path}"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        check=True,
    )
    return res.stdout.decode("utf-8")


def get_package_hash_from_latest_tag(package: str) -> str:
    """Get package hash from latest tag."""

    packages = json.loads(
        get_file_from_tag(
            file_path=str(Path(PACKAGES, PACKAGES_FILE)),
        )
    )

    return packages["dev"][package]



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the rounds of the skill."""

from copy import deepcopy

from packages.valory.skills.abstract_round_abci.base import BaseSynchronizedData
from packages.valory.skills.abstract_round_abci.test_tools.rounds import (
    BaseRoundTestClass as ExternalBaseRoundTestClass,
)
from packages.valory.skills.register_reset_recovery_abci.payloads import (
    RoundCountPayload,
)
from packages.valory.skills.register_reset_recovery_abci.rounds import (
    Event,
    RoundCountRound,
)


class BaseRoundTestClass(
    ExternalBaseRoundTestClass
):  # pylint: disable=too-few-public-methods
    """Base test class for Rounds."""

    _synchronized_data_class = BaseSynchronizedData
    _event_class = Event


class TestTerminationRound(BaseRoundTestClass):
    """Tests for TerminationRound."""

    def test_run(
        self,
    ) -> None:
        """Run tests."""

        test_round = RoundCountRound(
            synchronized_data=deepcopy(self.synchronized_data),
        )
        payload_data = 1
        first_payload, *payloads = [
            RoundCountPayload(sender=participant, current_round_count=payload_data)
            for participant in self.participants
        ]

        test_round.process_payload(first_payload)
        assert test_round.collection == {first_payload.sender: first_payload}
        assert test_round.end_block() is None

        for payload in payloads:
            test_round.process_payload(payload)

        res = test_round.end_block()
        assert res is not None
        _, event = res

        assert event == Event.DONE



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Integration tests for the valory/registration skill."""

# pylint: skip-file

import logging
import time
from pathlib import Path
from typing import Generator, Tuple

import pytest
from aea.configurations.data_types import PublicId
from aea_test_autonomy.base_test_classes.agents import (
    BaseTestEnd2End,
    BaseTestEnd2EndExecution,
    RoundChecks,
)
from aea_test_autonomy.fixture_helpers import (  # noqa: F401
    UseACNNode,
    UseRegistries,
    abci_host,
    abci_port,
    acn_config,
    acn_node,
    flask_tendermint,
    hardhat_port,
    ipfs_daemon,
    ipfs_domain,
    key_pairs,
    nb_nodes,
    registries_scope_class,
    tendermint_port,
    use_ipfs_daemon,
)

from packages.valory.skills.registration_abci.behaviours import (
    RegistrationStartupBehaviour,
)
from packages.valory.skills.registration_abci.rounds import RegistrationStartupRound


log_messages = RegistrationStartupBehaviour.LogMessages


@pytest.fixture(autouse=True)
def slow_down_tests() -> Generator:
    """Sleep in between tests"""
    logging.info("SLOWING DOWN TESTS")
    yield
    time.sleep(1)


# strict check log messages of the happy path
STRICT_CHECK_STRINGS = (
    log_messages.request_personal.value,
    log_messages.response_personal.value,
    log_messages.request_verification.value,
    log_messages.response_verification.value,
    log_messages.request_others.value,
    log_messages.collection_complete.value,
    log_messages.request_update.value,
    log_messages.response_update.value,
    "local height == remote height; continuing execution...",
)


HAPPY_PATH = (RoundChecks(RegistrationStartupRound.auto_round_id()),)


class RegistrationStartUpTestConfig(UseRegistries, UseACNNode, BaseTestEnd2End):
    """Base class for e2e tests using the ACN client connection"""

    skill_package = "valory/registration_abci:0.1.0"
    agent_package = "valory/registration_start_up:0.1.0"
    wait_to_finish = 60
    happy_path = HAPPY_PATH
    strict_check_strings: Tuple[str, ...] = STRICT_CHECK_STRINGS
    __p2p_prefix = "vendor.valory.connections.p2p_libp2p_client"
    __args_prefix = f"vendor.valory.skills.{PublicId.from_str(skill_package).name}.models.params.args"
    extra_configs = [
        {
            "dotted_path": f"{__args_prefix}.share_tm_config_on_startup",
            "value": True,
        },
        # setting the skill to non-abstract and setting a safe contract address is necessary to run an agent
        # We have set a null safe address in the `skill.yaml` because the safe address will not be utilized by the agent
        # We cannot set an override here, because of a limitation on the `open-aea`'s override mechanism that leads to:
        # Attribute `models.params.args.setup.safe_contract_address` is not allowed to be updated!
        # This exception is raised because the dict keys need to pre-exist in the config file in order to be overriden
        {
            "dotted_path": f"vendor.valory.skills.{PublicId.from_str(skill_package).name}.is_abstract",
            "value": False,
        },
        {
            "dotted_path": f"{__args_prefix}.reset_pause_duration",
            "value": 15,
        },
    ]

    def __set_configs(self, i: int, nb_agents: int) -> None:
        """Set the current agent's config overrides."""
        super().__set_configs(i=i, nb_agents=nb_agents)

        self.set_config(
            dotted_path=f"{self.__args_prefix}.tendermint_p2p_url",
            value=f"localhost:{self._tendermint_image.get_p2p_port(i=i)}",
            type_="str",
        )


@pytest.mark.e2e
@pytest.mark.integration
@pytest.mark.parametrize("nb_nodes", (4,))
class TestRegistrationStartUpFourAgents(
    RegistrationStartUpTestConfig, BaseTestEnd2EndExecution
):
    """Test registration start-up skill with four agents."""

    package_registry_src_rel = Path(__file__).parent.parent.parent.parent.parent


@pytest.mark.e2e
@pytest.mark.integration
@pytest.mark.parametrize("nb_nodes", (4,))
class TestRegistrationStartUpFourAgentsCatchUp(
    RegistrationStartUpTestConfig, BaseTestEnd2EndExecution
):
    """Test registration start-up skill with four agents and catch up."""

    stop_string = "My address: "
    restart_after = 10
    n_terminal = 1
    package_registry_src_rel = Path(__file__).parent.parent.parent.parent.parent



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Templates for FSM tests"""

from dataclasses import dataclass


@dataclass
class TEST_ROUNDS:
    """Template for test_rounds.py"""

    FILENAME = "test_rounds.py"

    HEADER = """\
    \"\"\"This package contains the tests for rounds of {FSMName}.\"\"\"

    from typing import Any, Type, Dict, List, Callable, Hashable, Mapping
    from dataclasses import dataclass, field

    import pytest

    from packages.{author}.skills.{skill_name}.payloads import (
        {payloads},
    )
    from packages.{author}.skills.{skill_name}.rounds import (
        AbstractRound,
        Event,
        SynchronizedData,
        {rounds},
    )
    from packages.valory.skills.abstract_round_abci.base import (
        BaseTxPayload,
    )
    from packages.valory.skills.abstract_round_abci.test_tools.rounds import (
        BaseRoundTestClass,
        BaseOnlyKeeperSendsRoundTest,
        BaseCollectDifferentUntilThresholdRoundTest,
        BaseCollectSameUntilThresholdRoundTest,
     )


    @dataclass
    class RoundTestCase:
        \"\"\"RoundTestCase\"\"\"

        name: str
        initial_data: Dict[str, Hashable]
        payloads: Mapping[str, BaseTxPayload]
        final_data: Dict[str, Hashable]
        event: Event
        synchronized_data_attr_checks: List[Callable] = field(default_factory=list)
        kwargs: Dict[str, Any] = field(default_factory=dict)


    MAX_PARTICIPANTS: int = 4

    """

    BASE_ROUND_TEST_CLS = """\
    class Base{FSMName}RoundTest(BaseRoundTestClass):
        \"\"\"Base test class for {FSMName} rounds.\"\"\"

        round_cls: Type[AbstractRound]
        synchronized_data: SynchronizedData
        _synchronized_data_class = SynchronizedData
        _event_class = Event

        def run_test(self, test_case: RoundTestCase) -> None:
            \"\"\"Run the test\"\"\"

            self.synchronized_data.update(**test_case.initial_data)

            test_round = self.round_cls(
                synchronized_data=self.synchronized_data,
            )

            self._complete_run(
                self._test_round(
                    test_round=test_round,
                    round_payloads=test_case.payloads,
                    synchronized_data_update_fn=lambda sync_data, _: sync_data.update(**test_case.final_data),
                    synchronized_data_attr_checks=test_case.synchronized_data_attr_checks,
                    exit_event=test_case.event,
                    **test_case.kwargs,  # varies per BaseRoundTestClass child
                )
            )

    """

    TEST_ROUND_CLS = """\
    class Test{RoundCls}(Base{FSMName}RoundTest):
        \"\"\"Tests for {RoundCls}.\"\"\"

        round_class = {RoundCls}

        # TODO: provide test cases
        @pytest.mark.parametrize("test_case", [])
        def test_run(self, test_case: RoundTestCase) -> None:
            \"\"\"Run tests.\"\"\"

            self.run_test(test_case)

    """


@dataclass
class TEST_BEHAVIOURS:
    """Template for test_behaviours.py"""

    FILENAME = "test_behaviours.py"

    HEADER = """\
    \"\"\"This package contains round behaviours of {AbciApp}.\"\"\"

    from pathlib import Path
    from typing import Any, Dict, Hashable, Optional, Type
    from dataclasses import dataclass, field

    import pytest

    from packages.valory.skills.abstract_round_abci.base import AbciAppDB
    from packages.valory.skills.abstract_round_abci.behaviours import (
        AbstractRoundBehaviour,
        BaseBehaviour,
        make_degenerate_behaviour,
    )
    from packages.{author}.skills.{skill_name}.behaviours import (
        {BaseBehaviourCls},
        {RoundBehaviourCls},
        {behaviours},
    )
    from packages.{author}.skills.{skill_name}.rounds import (
        SynchronizedData,
        DegenerateRound,
        Event,
        {AbciApp},
        {all_rounds},
    )

    from packages.valory.skills.abstract_round_abci.test_tools.base import (
        FSMBehaviourBaseCase,
    )


    @dataclass
    class BehaviourTestCase:
        \"\"\"BehaviourTestCase\"\"\"

        name: str
        initial_data: Dict[str, Hashable]
        event: Event
        kwargs: Dict[str, Any] = field(default_factory=dict)

    """

    BASE_BEHAVIOUR_TEST_CLS = """\
    class Base{FSMName}Test(FSMBehaviourBaseCase):
        \"\"\"Base test case.\"\"\"

        path_to_skill = Path(__file__).parent.parent

        behaviour: {RoundBehaviourCls}
        behaviour_class: Type[{BaseBehaviourCls}]
        next_behaviour_class: Type[{BaseBehaviourCls}]
        synchronized_data: SynchronizedData
        done_event = Event.DONE

        @property
        def current_behaviour_id(self) -> str:
            \"\"\"Current RoundBehaviour's behaviour id\"\"\"

            return self.behaviour.current_behaviour.behaviour_id

        def fast_forward(self, data: Optional[Dict[str, Any]] = None) -> None:
            \"\"\"Fast-forward on initialization\"\"\"

            data = data if data is not None else {{}}
            self.fast_forward_to_behaviour(
                self.behaviour,
                self.behaviour_class.behaviour_id,
                SynchronizedData(AbciAppDB(setup_data=AbciAppDB.data_to_lists(data))),
            )
            assert self.current_behaviour_id == self.behaviour_class.behaviour_id

        def complete(self, event: Event) -> None:
            \"\"\"Complete test\"\"\"

            self.behaviour.act_wrapper()
            self.mock_a2a_transaction()
            self._test_done_flag_set()
            self.end_round(done_event=event)
            assert self.current_behaviour_id == self.next_behaviour_class.behaviour_id

    """

    TEST_BEHAVIOUR_CLS = """\
    class Test{BehaviourCls}(Base{FSMName}Test):
        \"\"\"Tests {BehaviourCls}\"\"\"

        # TODO: set next_behaviour_class
        behaviour_class: Type[BaseBehaviour] = {BehaviourCls}
        next_behaviour_class: Type[BaseBehaviour] = ...

        # TODO: provide test cases
        @pytest.mark.parametrize("test_case", [])
        def test_run(self, test_case: BehaviourTestCase) -> None:
            \"\"\"Run tests.\"\"\"

            self.fast_forward(test_case.initial_data)
            # TODO: mock the necessary calls
            # self.mock_ ...
            self.complete(test_case.event)

    """


@dataclass
class TEST_PAYLOADS:
    """Template for test_payloads.py"""

    FILENAME = "test_payloads.py"

    HEADER = """\
    \"\"\"This package contains payload tests for the {AbciApp}.\"\"\"

    from typing import Type, Hashable
    from dataclasses import dataclass

    import pytest

    from packages.{author}.skills.{skill_name}.payloads import (
        BaseTxPayload,
        {payloads},
    )


    @dataclass
    class PayloadTestCase:
        \"\"\"PayloadTestCase\"\"\"

        name: str
        payload_cls: Type[BaseTxPayload]
        content: Hashable

    """

    TEST_PAYLOAD_CLS = """\
    # TODO: provide test cases
    @pytest.mark.parametrize("test_case", [])
    def test_payloads(test_case: PayloadTestCase) -> None:
        \"\"\"Tests for {AbciApp} payloads\"\"\"

        payload = test_case.payload_cls(sender="sender", content=test_case.content)
        assert payload.sender == "sender"
        assert payload.from_json(payload.json) == payload

    """


@dataclass
class TEST_MODELS:
    """Template for test_models.py"""

    FILENAME = "test_models.py"

    HEADER = """\
    \"\"\"Test the models.py module of the {FSMName}.\"\"\"

    from packages.valory.skills.abstract_round_abci.test_tools.base import DummyContext
    from packages.{author}.skills.{skill_name}.models import SharedState


    class TestSharedState:
        \"\"\"Test SharedState of {FSMName}.\"\"\"

        def test_initialization(self) -> None:
            \"\"\"Test initialization.\"\"\"
            SharedState(name="", skill_context=DummyContext())

    """


@dataclass
class TEST_HANDLERS:
    """Template for test_handlers.py"""

    FILENAME = "test_handlers.py"

    HEADER = """\
    \"\"\"Test the handlers.py module of the {FSMName}.\"\"\"

    import packages.{author}.skills.{skill_name}.handlers  # noqa


    def test_import() -> None:
        \"\"\"Test that the 'handlers.py' of the {FSMName} can be imported.\"\"\"

    """


@dataclass
class TEST_DIALOGUES:
    """Template for test_dialogues.py"""

    FILENAME = "test_dialogues.py"

    HEADER = """\
    \"\"\"Test the dialogues.py module of the {FSMName}.\"\"\"

    import packages.{author}.skills.{skill_name}.dialogues  # noqa


    def test_import() -> None:
        \"\"\"Test that the 'dialogues.py' of the {FSMName} can be imported.\"\"\"
    """



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test `valory/open-autonomy-tendermint` image."""

from pathlib import Path
from tempfile import TemporaryDirectory
from typing import cast

import docker
from aea_test_autonomy.docker.tendermint import DEFAULT_ABCI_HOST
from aea_test_autonomy.helpers.async_utils import wait_for_condition
from docker.models.containers import Container

from autonomy.constants import TENDERMINT_IMAGE_NAME, TENDERMINT_IMAGE_VERSION
from autonomy.deploy.constants import DOCKERFILES

from tests.conftest import ROOT_DIR, skip_docker_tests
from tests.test_autonomy.test_images.base import BaseImageBuildTest


@skip_docker_tests
class TestOpenAutonomyBaseImage(BaseImageBuildTest):
    """Test image build and run."""

    client: docker.DockerClient
    path: Path = ROOT_DIR / "deployments" / DOCKERFILES / "tendermint"
    tag: str = f"{TENDERMINT_IMAGE_NAME}:{TENDERMINT_IMAGE_VERSION}"

    def test_image(self) -> None:
        """Test image build."""

        # check build
        success, output = self.build_image(
            path=self.path,
            tag=self.tag,
        )

        assert success, output
        assert f"Successfully tagged {self.tag}" in output

        # check runtime

        with TemporaryDirectory() as temp_dir:
            log_file = Path(temp_dir, "logs.txt")
            log_file.touch()

            tm_container = cast(
                Container,
                self.client.containers.run(
                    image=self.tag,
                    detach=True,
                    network="host",
                    extra_hosts={
                        DEFAULT_ABCI_HOST: "host-gateway",
                    },
                    environment={
                        "TMHOME": "~/.tendermint",
                        "ID": "3",
                        "PROXY_APP": "tcp://localhost:26658",
                        "CREATE_EMPTY_BLOCKS": "false",
                        "LOG_FILE": "/logs/logs.txt",
                    },
                    volumes={
                        temp_dir: {"bind": "/logs", "mode": "rw"},
                    },
                ),
            )

            self.running_containers = [
                tm_container,
            ]

            def _check_for_outputs() -> bool:
                """Check for required outputs."""
                return "abci.socketClient failed to connect" in log_file.read_text()

            try:
                wait_for_condition(
                    condition_checker=_check_for_outputs,
                    timeout=10,
                    period=1,
                )
                successful = True
            except TimeoutError:
                successful = False

            assert successful, f"Tendermint runtime failed with {log_file.read_text()}"



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
# flake8: noqa: F811

"""Integration tests for the valory/register_reset skill."""

# pylint: skip-file

from pathlib import Path

import pytest
from aea.configurations.data_types import PublicId
from aea_test_autonomy.base_test_classes.agents import (
    BaseTestEnd2EndExecution,
    RoundChecks,
)
from aea_test_autonomy.configurations import KEY_PAIRS
from aea_test_autonomy.fixture_helpers import (  # noqa: F401
    UseACNNode,
    UseRegistries,
    abci_host,
    abci_port,
    acn_config,
    acn_node,
    flask_tendermint,
    hardhat_port,
    ipfs_daemon,
    ipfs_domain,
    key_pairs,
    nb_nodes,
    registries_scope_class,
    tendermint_port,
)

from packages.valory.skills.registration_abci.rounds import (
    RegistrationRound,
    RegistrationStartupRound,
)
from packages.valory.skills.reset_pause_abci.rounds import ResetAndPauseRound


HAPPY_PATH = (
    RoundChecks(RegistrationRound.auto_round_id(), n_periods=3),
    RoundChecks(ResetAndPauseRound.auto_round_id(), n_periods=4),
)


@pytest.mark.e2e
@pytest.mark.integration
@pytest.mark.flaky(reruns=1)
@pytest.mark.parametrize("nb_nodes", (4,))
class TestTendermintStartup(UseRegistries, UseACNNode, BaseTestEnd2EndExecution):
    """Test the ABCI register-reset skill with 4 agents starting up."""

    agent_package = "valory/register_reset:0.1.0"
    skill_package = "valory/register_reset_abci:0.1.0"
    happy_path = (
        RoundChecks(RegistrationStartupRound.auto_round_id()),
        RoundChecks(ResetAndPauseRound.auto_round_id()),
    )
    key_pairs = KEY_PAIRS
    wait_to_finish = 60
    package_registry_src_rel = Path(__file__).parent.parent.parent.parent.parent
    __args_prefix = f"vendor.valory.skills.{PublicId.from_str(skill_package).name}.models.params.args"
    extra_configs = [
        {
            "dotted_path": f"{__args_prefix}.reset_pause_duration",
            "value": 15,
        },
    ]


@pytest.mark.e2e
@pytest.mark.integration
@pytest.mark.parametrize("nb_nodes", (4,))
class TestTendermintReset(UseRegistries, UseACNNode, BaseTestEnd2EndExecution):
    """Test the ABCI register-reset skill with 4 agents when resetting Tendermint."""

    agent_package = "valory/register_reset:0.1.0"
    skill_package = "valory/register_reset_abci:0.1.0"
    happy_path = HAPPY_PATH
    key_pairs = KEY_PAIRS
    wait_to_finish = 200
    _reset_tendermint_every = 1
    _reset_pause_duration = 30
    package_registry_src_rel = Path(__file__).parent.parent.parent.parent.parent
    __args_prefix = f"vendor.valory.skills.{PublicId.from_str(skill_package).name}.models.params.args"
    # reset every `__reset_tendermint_every` rounds
    extra_configs = [
        {
            "dotted_path": f"{__args_prefix}.reset_tendermint_after",
            "value": _reset_tendermint_every,
        },
        {
            "dotted_path": f"{__args_prefix}.reset_pause_duration",
            "value": _reset_pause_duration,
        },
    ]


@pytest.mark.e2e
@pytest.mark.integration
@pytest.mark.flaky(reruns=1)
@pytest.mark.parametrize("nb_nodes", (4,))
class TestTendermintResetInterrupt(UseRegistries, UseACNNode, BaseTestEnd2EndExecution):
    """Test the ABCI register-reset skill with 4 agents when an agent gets interrupted on Tendermint reset."""

    agent_package = "valory/register_reset:0.1.0"
    skill_package = "valory/register_reset_abci:0.1.0"
    key_pairs = KEY_PAIRS
    wait_before_stop = 60
    wait_to_finish = 450
    restart_after = 1
    __reset_tendermint_every = 1
    stop_string = f"Entered in the '{ResetAndPauseRound.auto_round_id()}' round for period {__reset_tendermint_every - 1}"
    happy_path = HAPPY_PATH
    package_registry_src_rel = Path(__file__).parent.parent.parent.parent.parent
    __args_prefix = f"vendor.valory.skills.{PublicId.from_str(skill_package).name}.models.params.args"
    # reset every `__reset_tendermint_every` rounds
    extra_configs = [
        {
            "dotted_path": f"{__args_prefix}.reset_tendermint_after",
            "value": __reset_tendermint_every,
        },
        {
            "dotted_path": f"{__args_prefix}.reset_pause_duration",
            "value": 15,
        },
    ]

    n_terminal = 1


@pytest.mark.e2e
@pytest.mark.integration
@pytest.mark.flaky(reruns=1)
class TestTendermintResetInterruptNoRejoin(TestTendermintResetInterrupt):
    """
    Test a Tendermint reset case for the ABCI register-reset skill.

    Test the ABCI register-reset skill with 4 agents when an agent gets temporarily interrupted
    on Tendermint reset and never rejoins.
    """

    wait_to_finish = 200
    # set the restart to a value so that the agent never rejoins, in order to test the impact to the rest of the agents
    restart_after = wait_to_finish
    # check if we manage to reset with Tendermint with the rest of the agents; first agent will not rejoin in this test
    exclude_from_checks = [0]


@pytest.mark.e2e
@pytest.mark.integration
class TestTendermintResetRejoin(TestTendermintReset):
    """Test the ABCI register-reset skill with 4 agents when resetting Tendermint and one agent tries to rejoin."""

    stop_string = (
        f"Entered in the '{RegistrationRound.auto_round_id()}' round for period 1"
    )
    n_terminal = 1
    wait_before_stop = 300
    _n_resets_before_rejoin = 1
    restart_after = (
        TestTendermintReset._reset_pause_duration
        * TestTendermintReset._reset_tendermint_every
        * _n_resets_before_rejoin
    )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test push-all command group."""


import os
import shutil
from pathlib import Path
from typing import Any, Tuple
from unittest import mock

import _strptime  # noqa  # pylint: disable=unsed-import
from aea.cli.registry.settings import REMOTE_IPFS
from aea.configurations.constants import PACKAGES
from aea.package_manager.v1 import PackageManagerV1

from tests.conftest import ROOT_DIR
from tests.test_autonomy.test_cli.base import BaseCliTest


class TestPushAll(BaseCliTest):
    """Test `push-all` command."""

    cli_options: Tuple[str, ...] = ("push-all", "--remote")
    packages_dir: Path

    def setup(self) -> None:
        """Setup test method."""

        super().setup()

        self.packages_dir = self.t / PACKAGES
        shutil.copytree(ROOT_DIR / PACKAGES, self.packages_dir)
        os.chdir(self.t)

    def test_run(
        self,
    ) -> None:
        """Test run."""

        # packages/<author>/<component>/<name>/<config>.yaml
        available_packages = len(
            PackageManagerV1.from_dir(ROOT_DIR / PACKAGES).dev_packages
        )
        published_packages = []

        def _push_patch(_: Any, public_id: Any, retries: Any) -> None:
            published_packages.append(public_id)

        with mock.patch(
            "aea.cli.push_all.get_default_remote_registry", new=lambda: REMOTE_IPFS
        ), mock.patch("aea.cli.push_all.push_item_ipfs", new=_push_patch):
            result = self.run_cli()

        assert result.exit_code == 0, result.output
        assert len(published_packages) == available_packages



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for cli/helpers/fsm_spec.py"""

from pathlib import Path
from unittest import mock

import pytest
from _pytest.capture import CaptureFixture  # type: ignore
from click import ClickException

from autonomy.analyse.abci.app_spec import (
    DFA,
    DFASpecificationError,
    FSMSpecificationLoader,
)
from autonomy.cli.helpers.fsm_spec import (
    check_all,
    check_one,
    import_and_validate_app_class,
    update_one,
)

import packages
from packages.valory.skills import hello_world_abci, test_abci

from tests.conftest import ROOT_DIR


@pytest.mark.parametrize("relative_path", [True, False])
def test_import_and_validate_app_class(relative_path: bool) -> None:
    """Test import_and_validate_app_class"""

    module_path = Path(test_abci.__file__).parent
    if relative_path:
        module_path = module_path.relative_to(ROOT_DIR)
    module = import_and_validate_app_class(module_path, "TestAbciApp")
    assert module.__name__ == f"{test_abci.__name__}.rounds"


def test_import_and_validate_app_class_raises() -> None:
    """Test import and validate app class raises"""

    module_path = Path(packages.__file__).parent
    expected = "Cannot find the rounds module or the composition module for .*"
    with pytest.raises(FileNotFoundError, match=expected):
        import_and_validate_app_class(module_path, "DummyAbciApp")

    module_path = Path(test_abci.__file__).parent.relative_to(ROOT_DIR)
    expected = f'Class "DummyAbciApp" is not in "{test_abci.__name__}.rounds"'
    with pytest.raises(ClickException, match=expected):
        import_and_validate_app_class(module_path, "DummyAbciApp")


def test_update_one() -> None:
    """Test update_one"""

    package_path = Path(hello_world_abci.__file__).parent.relative_to(ROOT_DIR)
    with mock.patch.object(FSMSpecificationLoader, "dump") as m:
        update_one(package_path)
        m.assert_called_once()


def test_update_one_raises() -> None:
    """Test update_one raises"""

    package_path = Path(packages.__file__).parent
    expected = "FSM specification file .* does not exist, please provide app class name to continue."
    with pytest.raises(ClickException, match=expected):
        update_one(package_path)

    package_path = Path(hello_world_abci.__file__).parent.relative_to(ROOT_DIR)
    expected = "Please provide name for the app class or make sure FSM specification file is properly defined."
    with mock.patch.object(FSMSpecificationLoader, "load", return_value={}):
        with pytest.raises(ValueError, match=expected):
            update_one(package_path)


def test_check_one() -> None:
    """Test check_one"""

    package_path = Path(hello_world_abci.__file__).parent.relative_to(ROOT_DIR)
    check_one(package_path)


def test_check_one_raises() -> None:
    """Test check_one raises"""

    package_path = Path(hello_world_abci.__file__).parent.relative_to(ROOT_DIR)

    expected = "Please provide name for the app class or make sure FSM specification file is properly defined."
    with mock.patch.object(FSMSpecificationLoader, "load", return_value={}):
        with pytest.raises(ValueError, match=expected):
            update_one(package_path)

    expected = 'Class .* is not in "packages.valory.skills.hello_world_abci.rounds"'
    with pytest.raises(ClickException, match=expected):
        check_one(package_path, app_class="DummyAbciApp")

    target = "autonomy.cli.helpers.fsm_spec.check_unreferenced_events"
    with mock.patch(target, return_value=["Event.WIN_LOTTERY"]):
        expected = "Event reference check failed with .*"
        with pytest.raises(DFASpecificationError, match=expected):
            check_one(package_path)

    with mock.patch.object(DFA, "load", return_value=""):
        expected = "FSM Spec definition does not match in specification file and class definitions."
        with pytest.raises(DFASpecificationError, match=expected):
            check_one(package_path)


@pytest.mark.parametrize("relative_path", [True, False])
def test_check_all(relative_path: bool, capsys: CaptureFixture) -> None:
    """Test check_all"""

    packages_dir = Path(packages.__file__).parent
    if relative_path:
        packages_dir = packages_dir.relative_to(ROOT_DIR)

    # there currently exists several fsm_specification.yaml files
    with mock.patch("autonomy.cli.helpers.fsm_spec.check_one") as m:
        check_all(packages_dir, FSMSpecificationLoader.OutputFormats.YAML)
        m.assert_called()
        captured = capsys.readouterr()
        assert "Checking" in captured.out

    # no fsm_specification.json files
    with mock.patch("autonomy.cli.helpers.fsm_spec.check_one") as m:
        check_all(packages_dir, FSMSpecificationLoader.OutputFormats.JSON)
        m.assert_not_called()
        captured = capsys.readouterr()
        assert "Checking" not in captured.out

    # no fsm_specification.mermaid files
    with mock.patch("autonomy.cli.helpers.fsm_spec.check_one") as m:
        check_all(packages_dir, FSMSpecificationLoader.OutputFormats.MERMAID)
        m.assert_not_called()
        captured = capsys.readouterr()
        assert "Checking" not in captured.out


def test_check_all_raises() -> None:
    """Test check_all raises"""

    package_dir = Path(packages.__file__).parent.relative_to(ROOT_DIR)

    target = "autonomy.cli.helpers.fsm_spec.check_one"
    with mock.patch(target, side_effect=DFASpecificationError):
        expected = "Specifications check for following packages failed."
        with pytest.raises(DFASpecificationError, match=expected):
            check_all(package_dir)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""ABCI App test tools."""


from abc import ABC
from enum import Enum
from typing import Dict, Tuple, Type, Union
from unittest.mock import MagicMock

from packages.valory.skills.abstract_round_abci.base import (
    AbciApp,
    AbstractRound,
    BaseSynchronizedData,
    BaseTxPayload,
)


class _ConcreteRound(AbstractRound, ABC):
    """ConcreteRound"""

    synchronized_data_class = BaseSynchronizedData
    payload_attribute = ""

    def end_block(self) -> Union[None, Tuple[MagicMock, MagicMock]]:
        """End block."""

    def check_payload(self, payload: BaseTxPayload) -> None:
        """Check payload."""

    def process_payload(self, payload: BaseTxPayload) -> None:
        """Process payload."""


class ConcreteRoundA(_ConcreteRound):
    """Dummy instantiation of the AbstractRound class."""

    payload_class = BaseTxPayload

    def end_block(self) -> Tuple[MagicMock, MagicMock]:
        """End block."""
        return MagicMock(), MagicMock()


class ConcreteRoundB(_ConcreteRound):
    """Dummy instantiation of the AbstractRound class."""

    payload_class = BaseTxPayload


class ConcreteRoundC(_ConcreteRound):
    """Dummy instantiation of the AbstractRound class."""

    payload_class = BaseTxPayload


class ConcreteBackgroundRound(_ConcreteRound):
    """Dummy instantiation of the AbstractRound class."""

    payload_class = BaseTxPayload


class ConcreteTerminationRoundA(_ConcreteRound):
    """Dummy instantiation of the AbstractRound class."""

    payload_class = BaseTxPayload


class ConcreteTerminationRoundB(_ConcreteRound):
    """Dummy instantiation of the AbstractRound class."""

    payload_class = BaseTxPayload


class ConcreteTerminationRoundC(_ConcreteRound):
    """Dummy instantiation of the AbstractRound class."""

    payload_class = BaseTxPayload


class ConcreteEvents(Enum):
    """Defines dummy events to be used for testing purposes."""

    TERMINATE = "terminate"
    A = "a"
    B = "b"
    C = "c"
    TIMEOUT = "timeout"

    def __str__(self) -> str:
        """Get the string representation of the event."""
        return self.value


class AbciAppTest(AbciApp[ConcreteEvents]):
    """A dummy AbciApp for testing purposes."""

    TIMEOUT: float = 1.0

    initial_round_cls: Type[AbstractRound] = ConcreteRoundA
    transition_function: Dict[
        Type[AbstractRound], Dict[ConcreteEvents, Type[AbstractRound]]
    ] = {
        ConcreteRoundA: {
            ConcreteEvents.A: ConcreteRoundA,
            ConcreteEvents.B: ConcreteRoundB,
            ConcreteEvents.C: ConcreteRoundC,
        },
        ConcreteRoundB: {
            ConcreteEvents.B: ConcreteRoundB,
            ConcreteEvents.TIMEOUT: ConcreteRoundA,
        },
        ConcreteRoundC: {
            ConcreteEvents.C: ConcreteRoundA,
            ConcreteEvents.TIMEOUT: ConcreteRoundC,
        },
    }
    background_round_cls = ConcreteBackgroundRound
    termination_transition_function: Dict[
        Type[AbstractRound], Dict[ConcreteEvents, Type[AbstractRound]]
    ] = {
        ConcreteBackgroundRound: {
            ConcreteEvents.TERMINATE: ConcreteTerminationRoundA,
        },
        ConcreteTerminationRoundA: {
            ConcreteEvents.A: ConcreteTerminationRoundA,
            ConcreteEvents.B: ConcreteTerminationRoundB,
            ConcreteEvents.C: ConcreteTerminationRoundC,
        },
        ConcreteTerminationRoundB: {
            ConcreteEvents.B: ConcreteTerminationRoundB,
            ConcreteEvents.TIMEOUT: ConcreteTerminationRoundA,
        },
        ConcreteTerminationRoundC: {
            ConcreteEvents.C: ConcreteTerminationRoundA,
            ConcreteEvents.TIMEOUT: ConcreteTerminationRoundC,
        },
    }
    termination_event = ConcreteEvents.TERMINATE
    event_to_timeout: Dict[ConcreteEvents, float] = {
        ConcreteEvents.TIMEOUT: TIMEOUT,
    }



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test base configurations."""

import os
import re
import shutil
import tempfile
from pathlib import Path
from typing import Dict, List

import pytest
import yaml
from aea.configurations.data_types import PublicId
from aea.exceptions import AEAValidationError
from aea.helpers.io import open_file
from aea.helpers.yaml_utils import yaml_load_all

from autonomy.configurations.base import Service
from autonomy.configurations.loader import load_service_config

from tests.conftest import ROOT_DIR
from tests.test_autonomy.base import get_dummy_service_config


class TestServiceConfig:
    """Test `Service` component."""

    cwd: Path
    t: Path
    service_file: Path

    @classmethod
    def setup_class(
        cls,
    ) -> None:
        """Setup test class."""
        cls.cwd = Path.cwd()

    def setup(
        self,
    ) -> None:
        """Setup test."""
        self.t = Path(tempfile.TemporaryDirectory().name)
        self.t.mkdir()
        self.service_file = self.t / "service.yaml"

        os.chdir(self.t)

    def _write_service(self, data: List[Dict]) -> None:
        """Write service config to a file."""
        with open(self.service_file, "w+") as fp:
            yaml.dump_all(data, fp)

    def test_load_service(
        self,
    ) -> None:
        """Test load service."""
        dummy_service = get_dummy_service_config()
        self._write_service(dummy_service)

        service = load_service_config(self.t)
        for key, value in service.json.items():
            if key == "overrides":
                continue
            assert value == dummy_service[0][key]

    def test_check_overrides_valid_fail_missing_overrides(
        self,
    ) -> None:
        """Test check_overrides_valid method."""

        dummy_service = get_dummy_service_config(file_number=2)
        dummy_service[0]["number_of_agents"] = 6
        self._write_service(dummy_service)

        with pytest.raises(
            AEAValidationError,
            match=re.escape(
                "Not enough overrides for component (skill, valory/dummy_skill:0.1.0); Number of agents: 6"
            ),
        ):
            load_service_config(self.t)

    def test_check_repeatetive_overrides(
        self,
    ) -> None:
        """Test check_overrides_valid method."""

        dummy_service = get_dummy_service_config(file_number=2)
        (skill_config,) = [
            override for override in dummy_service if override.get("type") == "skill"
        ]
        dummy_service.append(skill_config)
        self._write_service(dummy_service)

        with pytest.raises(
            AEAValidationError,
            match=re.escape(
                "Overrides for component (skill, valory/dummy_skill:0.1.0) are defined more than once"
            ),
        ):
            load_service_config(self.t)

    def test_process_metadata(
        self,
    ) -> None:
        """Test process metadata."""

        _, component_override, _ = get_dummy_service_config(file_number=2)
        configuration, component_id, has_multiple_overrides = Service.process_metadata(
            component_override.copy()
        )

        assert has_multiple_overrides
        assert "public_id" not in configuration
        assert "type" not in configuration
        assert (
            PublicId.from_str(component_override["public_id"]) == component_id.public_id
        )

    def test_no_configuration_provided_for_agent_x(
        self,
    ) -> None:
        """Test process metadata."""

        service = Service(name="service", author="author", agent="valory/agent")
        with pytest.raises(
            ValueError,
            match=re.escape(
                "Overrides not provided for agent 1; component=(skill, valory/skill:latest)"
            ),
        ):
            service.process_component_overrides(
                1, {"type": "skill", "public_id": "valory/skill"}
            )

    def test_env_vars(
        self,
    ) -> None:
        """Test if env vars are properly loaded."""

        env_placeholder = "${NUMBER_OF_AGENTS:int:1}"
        dummy_service = get_dummy_service_config()
        dummy_service[0]["number_of_agents"] = env_placeholder

        self._write_service(dummy_service)

        service = load_service_config(self.t, substitute_env_vars=True)
        assert service.number_of_agents == 1

    def teardown(
        self,
    ) -> None:
        """Teardown test."""
        os.chdir(self.cwd)
        shutil.rmtree(self.t)


@pytest.mark.parametrize(
    "service_file",
    Path(
        ROOT_DIR,
        "tests",
        "data",
        "dummy_service_config_files",
    ).iterdir(),
)
def test_load_service(service_file: Path) -> None:
    """Test loading and processing service component."""

    with open_file(service_file, "r", encoding="utf-8") as fp:
        data = yaml_load_all(fp)

    service_config, *overrides = data
    Service.validate_config_data(service_config)
    service_config["license_"] = service_config.pop("license")

    service = Service(**service_config)
    service.overrides = overrides



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Develop CLI module."""


import click
from docker import from_env

from autonomy.constants import (
    DEFAULT_SERVICE_REGISTRY_CONTRACTS_IMAGE,
    SERVICE_REGISTRY_CONTRACT_CONTAINER_NAME,
)


@click.group(name="develop")
def develop_group() -> None:
    """Develop an agent service."""


@develop_group.command(name="service-registry-network")
@click.argument(
    "image",
    type=str,
    required=False,
    default=DEFAULT_SERVICE_REGISTRY_CONTRACTS_IMAGE,
)
def run_service_locally(image: str) -> None:
    """Run the service registry contracts on a local network."""
    click.echo(f"Starting {image}.")
    client = from_env()
    container = client.containers.run(
        image=image,
        detach=True,
        network_mode="host",
        name=SERVICE_REGISTRY_CONTRACT_CONTAINER_NAME,
    )
    try:
        for line in client.api.logs(container.id, follow=True, stream=True):
            click.echo(line.decode())
    except KeyboardInterrupt:  # pragma: no cover
        pass
    finally:  # pragma: no cover
        click.echo("Stopping container.")
        container.stop()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Deploy CLI module."""

import shutil
from pathlib import Path
from typing import Optional, cast

import click
from aea.cli.registry.settings import REGISTRY_REMOTE
from aea.cli.utils.click_utils import (
    password_option,
    registry_flag,
    reraise_as_click_exception,
)
from aea.cli.utils.context import Context
from aea.configurations.constants import DEFAULT_ENV_DOTFILE

from autonomy.chain.config import ChainType
from autonomy.cli.helpers.deployment import (
    build_and_deploy_from_token,
    build_deployment,
    run_deployment,
)
from autonomy.cli.helpers.env import load_env_file
from autonomy.cli.utils.click_utils import (
    PathArgument,
    chain_selection_flag,
    image_author_option,
)
from autonomy.constants import DEFAULT_BUILD_FOLDER, DEFAULT_KEYS_FILE
from autonomy.deploy.base import NotValidKeysFile
from autonomy.deploy.constants import INFO, LOGGING_LEVELS
from autonomy.deploy.generators.docker_compose.base import DockerComposeGenerator
from autonomy.deploy.generators.kubernetes.base import KubernetesGenerator


PACKAGES_DIR = "packages_dir"
OPEN_AEA_DIR = "open_aea_dir"
OPEN_AUTONOMY_DIR = "open_autonomy_dir"


@click.group(name="deploy")
@click.option(
    "--env-file",
    type=PathArgument(
        exists=True,
        dir_okay=False,
        file_okay=True,
    ),
    help="File containing environment variable mappings",
)
@click.pass_context
def deploy_group(
    click_context: click.Context,  # pylint: disable=unused-argument
    env_file: Optional[Path],
) -> None:
    """Deploy an agent service."""
    dot_env_file = Path.cwd() / DEFAULT_ENV_DOTFILE
    if dot_env_file.exists():
        load_env_file(file=dot_env_file)
    if env_file is not None:
        load_env_file(file=env_file, serialize_json=True)


@deploy_group.command(name="build")
@click.argument("keys_file", type=str, required=False)
@click.option(
    "--o",
    "output_dir",
    type=click.Path(exists=False, dir_okay=True),
    help="Path to output dir.",
)
@click.option(
    "--n",
    "number_of_agents",
    type=int,
    default=None,
    help="Number of agents.",
)
@click.option(
    "--docker",
    "deployment_type",
    flag_value=DockerComposeGenerator.deployment_type,
    default=True,
    help="Use docker as a backend.",
)
@click.option(
    "--kubernetes",
    "deployment_type",
    flag_value=KubernetesGenerator.deployment_type,
    help="Use kubernetes as a backend.",
)
@click.option(
    "--dev",
    "dev_mode",
    is_flag=True,
    default=False,
    help="Create development environment.",
)
@click.option(
    "--log-level",
    type=click.Choice(choices=LOGGING_LEVELS, case_sensitive=True),
    help="Logging level for runtime.",
    default=INFO,
)
@click.option(
    "--packages-dir", type=click.Path(), help="Path to packages dir (Use with dev mode)"
)
@click.option(
    "--open-aea-dir",
    type=click.Path(),
    help="Path to open-aea repo (Use with dev mode)",
)
@click.option(
    "--open-autonomy-dir",
    type=click.Path(),
    help="Path to open-autonomy repo (Use with dev mode)",
)
@click.option(
    "--aev",
    is_flag=True,
    default=False,
    help="Apply environment variable when loading service config.",
)
@click.option(
    "--use-hardhat",
    is_flag=True,
    default=False,
    help="Include a hardhat node in the deployment setup.",
)
@click.option(
    "--use-acn",
    is_flag=True,
    default=False,
    help="Include an ACN node in the deployment setup.",
)
@click.option(
    "-ltm",
    "--local-tm-setup",
    "use_tm_testnet_setup",
    is_flag=True,
    default=False,
    help="Use local tendermint chain setup.",
)
@click.option("--image-version", type=str, help="Define runtime image version.")
@registry_flag()
@password_option(confirmation_prompt=True)
@image_author_option
@click.pass_context
def build_deployment_command(  # pylint: disable=too-many-arguments, too-many-locals
    click_context: click.Context,
    keys_file: Optional[Path],
    deployment_type: str,
    output_dir: Optional[Path],
    dev_mode: bool,
    registry: str,
    number_of_agents: Optional[int] = None,
    password: Optional[str] = None,
    open_aea_dir: Optional[Path] = None,
    packages_dir: Optional[Path] = None,
    open_autonomy_dir: Optional[Path] = None,
    log_level: str = INFO,
    aev: bool = False,
    image_version: Optional[str] = None,
    use_hardhat: bool = False,
    use_acn: bool = False,
    use_tm_testnet_setup: bool = False,
    image_author: Optional[str] = None,
) -> None:
    """Build deployment setup for n agents."""

    keys_file = Path(keys_file or DEFAULT_KEYS_FILE).absolute()
    if not keys_file.exists():
        message = f"No such file or directory: {keys_file}. Please provide valid path for keys file."
        raise click.ClickException(message)

    build_dir = Path(output_dir or DEFAULT_BUILD_FOLDER).absolute()
    packages_dir = Path(packages_dir or Path.cwd() / "packages").absolute()
    open_aea_dir = Path(open_aea_dir or Path.home() / "open-aea").absolute()
    open_autonomy_dir = Path(
        open_autonomy_dir or Path.home() / "open-autonomy"
    ).absolute()

    if dev_mode:
        for name, path in (
            (PACKAGES_DIR, packages_dir),
            (OPEN_AEA_DIR, open_aea_dir),
            (OPEN_AUTONOMY_DIR, open_autonomy_dir),
        ):
            if not path.exists():
                flag = "--" + "-".join(name.split("_"))
                raise click.ClickException(
                    f"Path does not exist @ {path} for {name}; Please provide proper value for {flag}"
                )

    ctx = cast(Context, click_context.obj)
    ctx.registry_type = registry

    try:
        build_deployment(
            keys_file=keys_file,
            build_dir=build_dir,
            deployment_type=deployment_type,
            dev_mode=dev_mode,
            number_of_agents=number_of_agents,
            password=password,
            packages_dir=packages_dir,
            open_aea_dir=open_aea_dir,
            open_autonomy_dir=open_autonomy_dir,
            log_level=log_level,
            apply_environment_variables=aev,
            image_version=image_version,
            use_hardhat=use_hardhat,
            use_acn=use_acn,
            use_tm_testnet_setup=use_tm_testnet_setup,
            image_author=image_author,
        )
    except (NotValidKeysFile, FileNotFoundError, FileExistsError) as e:
        shutil.rmtree(build_dir)
        raise click.ClickException(str(e)) from e


@deploy_group.command(name="run")
@click.option(
    "--build-dir",
    type=click.Path(),
)
@click.option(
    "--no-recreate",
    is_flag=True,
    default=False,
    help="If containers already exist, don't recreate them.",
)
@click.option(
    "--remove-orphans",
    is_flag=True,
    default=False,
    help="Remove containers for services not defined in the Compose file.",
)
def run(build_dir: Path, no_recreate: bool, remove_orphans: bool) -> None:
    """Run deployment."""
    build_dir = Path(build_dir or Path.cwd()).absolute()

    if not (build_dir / DockerComposeGenerator.output_name).exists():
        raise click.ClickException(
            f"Deployment configuration does not exist @ {build_dir}"
        )

    run_deployment(build_dir, no_recreate, remove_orphans)


@deploy_group.command(name="from-token")
@click.argument("token_id", type=int)
@click.argument("keys_file", type=click.Path())
@click.option("--n", type=int, help="Number of agents to include in the build.")
@click.option("--skip-image", is_flag=True, default=False, help="Skip building images.")
@click.option(
    "--aev",
    is_flag=True,
    default=False,
    help="Apply environment variable when loading service config.",
)
@click.option(
    "--docker",
    "deployment_type",
    flag_value=DockerComposeGenerator.deployment_type,
    default=True,
    help="Use docker as a backend.",
)
@click.option(
    "--kubernetes",
    "deployment_type",
    flag_value=KubernetesGenerator.deployment_type,
    help="Use kubernetes as a backend.",
)
@click.option(
    "--no-deploy",
    is_flag=True,
    help="If set to true, the deployment won't run automatically",
)
@chain_selection_flag(help_string_format="Use {} chain to resolve the token id.")
@click.pass_context
@password_option(confirmation_prompt=True)
def run_deployment_from_token(  # pylint: disable=too-many-arguments, too-many-locals
    click_context: click.Context,
    token_id: int,
    keys_file: Path,
    chain_type: ChainType,
    skip_image: bool,
    n: Optional[int],
    deployment_type: str,
    no_deploy: bool,
    aev: bool = False,
    password: Optional[str] = None,
) -> None:
    """Run service deployment."""

    ctx = cast(Context, click_context.obj)
    ctx.registry_type = REGISTRY_REMOTE
    keys_file = Path(keys_file or DEFAULT_KEYS_FILE).absolute()

    with reraise_as_click_exception(
        NotValidKeysFile, FileNotFoundError, FileExistsError
    ):
        build_and_deploy_from_token(
            token_id=token_id,
            keys_file=keys_file,
            chain_type=ChainType(chain_type),
            skip_image=skip_image,
            n=n,
            deployment_type=deployment_type,
            aev=aev,
            password=password,
            no_deploy=no_deploy,
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This package contains round behaviours of DummyAbciApp."""

from abc import ABC
from collections import deque
from typing import Deque, Generator, Set, Type, cast

from packages.valory.skills.abstract_round_abci.base import AbstractRound
from packages.valory.skills.abstract_round_abci.behaviours import (
    AbstractRoundBehaviour,
    BaseBehaviour,
)
from packages.valory.skills.abstract_round_abci.common import (
    RandomnessBehaviour,
    SelectKeeperBehaviour,
)
from packages.valory.skills.abstract_round_abci.tests.data.dummy_abci.models import (
    Params,
)
from packages.valory.skills.abstract_round_abci.tests.data.dummy_abci.payloads import (
    DummyFinalPayload,
    DummyKeeperSelectionPayload,
    DummyRandomnessPayload,
    DummyStartingPayload,
)
from packages.valory.skills.abstract_round_abci.tests.data.dummy_abci.rounds import (
    DummyAbciApp,
    DummyFinalRound,
    DummyKeeperSelectionRound,
    DummyRandomnessRound,
    DummyStartingRound,
    SynchronizedData,
)


class DummyBaseBehaviour(BaseBehaviour, ABC):
    """Base behaviour for the common apps' skill."""

    @property
    def synchronized_data(self) -> SynchronizedData:
        """Return the synchronized data."""
        return cast(SynchronizedData, super().synchronized_data)

    @property
    def params(self) -> Params:
        """Return the params."""
        return cast(Params, super().params)


class DummyStartingBehaviour(DummyBaseBehaviour):
    """DummyStartingBehaviour"""

    behaviour_id: str = "dummy_starting"
    matching_round: Type[AbstractRound] = DummyStartingRound

    def async_act(self) -> Generator:
        """Do the act, supporting asynchronous execution."""

        with self.context.benchmark_tool.measure(self.behaviour_id).local():
            content = "dummy"
            sender = self.context.agent_address
            payload = DummyStartingPayload(sender=sender, content=content)

        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            yield from self.wait_until_round_end()

        self.set_done()


class DummyRandomnessBehaviour(RandomnessBehaviour):
    """DummyRandomnessBehaviour"""

    behaviour_id: str = "dummy_randomness"
    matching_round: Type[AbstractRound] = DummyRandomnessRound
    payload_class = DummyRandomnessPayload


class DummyKeeperSelectionBehaviour(SelectKeeperBehaviour):
    """DummyKeeperSelectionBehaviour"""

    behaviour_id: str = "dummy_keeper_selection"
    matching_round: Type[AbstractRound] = DummyKeeperSelectionRound
    payload_class = DummyKeeperSelectionPayload

    @staticmethod
    def serialized_keepers(keepers: Deque[str], keeper_retries: int = 1) -> str:
        """Get the keepers serialized."""
        if not keepers:
            return ""
        return keeper_retries.to_bytes(32, "big").hex() + "".join(keepers)

    def async_act(self) -> Generator:
        """Do the act, supporting asynchronous execution."""

        with self.context.benchmark_tool.measure(self.behaviour_id).local():
            keepers = deque((self._select_keeper(),))
            payload = self.payload_class(
                self.context.agent_address, self.serialized_keepers(keepers)
            )

        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            yield from self.wait_until_round_end()

        self.set_done()


class DummyFinalBehaviour(DummyBaseBehaviour):
    """DummyFinalBehaviour"""

    behaviour_id: str = "dummy_final"
    matching_round: Type[AbstractRound] = DummyFinalRound

    def async_act(self) -> Generator:
        """Do the act, supporting asynchronous execution."""

        with self.context.benchmark_tool.measure(self.behaviour_id).local():
            content = True
            sender = self.context.agent_address
            payload = DummyFinalPayload(sender=sender, content=content)

        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            yield from self.wait_until_round_end()

        self.set_done()


class DummyRoundBehaviour(AbstractRoundBehaviour):
    """DummyRoundBehaviour"""

    initial_behaviour_cls = DummyStartingBehaviour
    abci_app_cls = DummyAbciApp
    behaviours: Set[Type[BaseBehaviour]] = {
        DummyFinalBehaviour,  # type: ignore
        DummyKeeperSelectionBehaviour,  # type: ignore
        DummyRandomnessBehaviour,  # type: ignore
        DummyStartingBehaviour,  # type: ignore
    }



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Integration tests for the valory/register_reset skill."""


from pathlib import Path

import pytest
from aea.configurations.data_types import PublicId
from aea_test_autonomy.base_test_classes.agents import RoundChecks
from aea_test_autonomy.fixture_helpers import (  # noqa: F401  # pylint: disable=unused-import
    UseACNNode,
    UseRegistries,
    abci_host,
    abci_port,
    acn_config,
    acn_node,
    flask_tendermint,
    hardhat_port,
    ipfs_daemon,
    ipfs_domain,
    key_pairs,
    nb_nodes,
    registries_scope_class,
    tendermint_port,
)

from packages.valory.agents.register_reset_recovery.tests.base import (
    BaseTestRegisterResetRecoveryEnd2End,
)
from packages.valory.skills.registration_abci.rounds import RegistrationStartupRound


HAPPY_PATH = (RoundChecks(RegistrationStartupRound.auto_round_id()),)

# the string used to trigger the breaking of the 3rd tm node (node3).
TM_BREAK_STRING = "Current round count is 3."

# we make sure that all agents are able to reach round 25.
# the third one will communication wth tendermint at round 3.
STRICT_CHECK_STRINGS = ("Current round count is 25.",)


@pytest.mark.e2e
@pytest.mark.integration
@pytest.mark.parametrize("nb_nodes", (4,))
class TestRegisterResetRecoverStartup(
    UseRegistries, UseACNNode, BaseTestRegisterResetRecoveryEnd2End
):
    """Test the ABCI register-reset skill with 4 agents starting up."""

    agent_package = "valory/register_reset_recovery:0.1.0"
    skill_package = "valory/register_reset_recovery_abci:0.1.0"
    happy_path = HAPPY_PATH
    tm_break_string = TM_BREAK_STRING
    strict_check_strings = STRICT_CHECK_STRINGS
    wait_to_finish = 300
    package_registry_src_rel = Path(__file__).parents[4]
    __args_prefix = f"vendor.valory.skills.{PublicId.from_str(skill_package).name}.models.params.args"
    extra_configs = [
        {
            "dotted_path": f"{__args_prefix}.reset_pause_duration",
            "value": 15,
        },
    ]



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This package contains the rounds of DummyAbciApp."""

from abc import ABC
from enum import Enum
from typing import FrozenSet, Optional, Set, Tuple, cast

from packages.valory.skills.abstract_round_abci.base import (
    AbciApp,
    AbciAppTransitionFunction,
    AbstractRound,
    AppState,
    BaseSynchronizedData,
    CollectSameUntilAllRound,
    CollectSameUntilThresholdRound,
    EventToTimeout,
    OnlyKeeperSendsRound,
)
from packages.valory.skills.abstract_round_abci.tests.data.dummy_abci.payloads import (
    DummyFinalPayload,
    DummyKeeperSelectionPayload,
    DummyRandomnessPayload,
    DummyStartingPayload,
)


class Event(Enum):
    """DummyAbciApp Events"""

    ROUND_TIMEOUT = "round_timeout"
    NO_MAJORITY = "no_majority"
    DONE = "done"


class SynchronizedData(BaseSynchronizedData):
    """
    Class to represent the synchronized data.

    This data is replicated by the tendermint application.
    """


class DummyMixinRound(AbstractRound, ABC):
    """DummyMixinRound"""

    done_event = Event.DONE
    no_majority_event = Event.NO_MAJORITY

    @property
    def synchronized_data(self) -> SynchronizedData:
        """Return the synchronized data."""
        return cast(SynchronizedData, self._synchronized_data)


class DummyStartingRound(CollectSameUntilAllRound, DummyMixinRound):
    """DummyStartingRound"""

    round_id: str = "dummy_starting"
    payload_class = DummyStartingPayload
    payload_attribute: str = "dummy_starting"
    synchronized_data_class = SynchronizedData

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Event]]:
        """Process the end of the block."""

        if self.collection_threshold_reached:
            synchronized_data = self.synchronized_data.update(
                participants=tuple(sorted(self.collection)),
                synchronized_data_class=SynchronizedData,
            )
            return synchronized_data, Event.DONE
        return None


class DummyRandomnessRound(CollectSameUntilThresholdRound, DummyMixinRound):
    """DummyRandomnessRound"""

    round_id: str = "dummy_randomness"
    payload_class = DummyRandomnessPayload
    payload_attribute: str = "dummy_randomness"
    collection_key = "participant_to_randomness"
    selection_key = "most_voted_randomness"
    synchronized_data_class = SynchronizedData


class DummyKeeperSelectionRound(CollectSameUntilThresholdRound, DummyMixinRound):
    """DummyKeeperSelectionRound"""

    round_id: str = "dummy_keeper_selection"
    payload_class = DummyKeeperSelectionPayload
    payload_attribute: str = "dummy_keeper_selection"
    collection_key = "participant_to_keeper"
    selection_key = "most_voted_keeper"
    synchronized_data_class = SynchronizedData


class DummyFinalRound(OnlyKeeperSendsRound, DummyMixinRound):
    """DummyFinalRound"""

    round_id: str = "dummy_final"
    payload_class = DummyFinalPayload
    payload_attribute: str = "dummy_final"
    synchronized_data_class = SynchronizedData


class DummyAbciApp(AbciApp[Event]):
    """DummyAbciApp"""

    initial_round_cls: AppState = DummyStartingRound
    transition_function: AbciAppTransitionFunction = {
        DummyStartingRound: {
            Event.DONE: DummyRandomnessRound,
            Event.ROUND_TIMEOUT: DummyStartingRound,
            Event.NO_MAJORITY: DummyStartingRound,
        },
        DummyRandomnessRound: {
            Event.DONE: DummyKeeperSelectionRound,
            Event.ROUND_TIMEOUT: DummyRandomnessRound,
            Event.NO_MAJORITY: DummyRandomnessRound,
        },
        DummyKeeperSelectionRound: {
            Event.DONE: DummyFinalRound,
            Event.ROUND_TIMEOUT: DummyKeeperSelectionRound,
            Event.NO_MAJORITY: DummyKeeperSelectionRound,
        },
        DummyFinalRound: {
            Event.DONE: DummyStartingRound,
            Event.ROUND_TIMEOUT: DummyFinalRound,
            Event.NO_MAJORITY: DummyFinalRound,
        },
    }
    final_states: Set[AppState] = set()
    event_to_timeout: EventToTimeout = {}
    cross_period_persisted_keys: FrozenSet[str] = frozenset()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
# pylint: disable=broad-except,unspecified-encoding,import-error,redefined-outer-name

"""End2End tests base classes for the Termination agent."""
import web3
from aea.configurations.base import PublicId
from aea_test_autonomy.base_test_classes.agents import BaseTestEnd2End
from aea_test_autonomy.configurations import KEY_PAIRS as _HARDHAT_KEY_PAIRS
from aea_test_autonomy.docker.registries import (
    DEFAULT_HARDHAT_ADDR as _DEFAULT_HARDHAT_ADDR,
)
from aea_test_autonomy.docker.registries import (
    DEFAULT_HARDHAT_PORT as _DEFAULT_HARDHAT_PORT,
)
from aea_test_autonomy.docker.registries import (
    GNOSIS_SAFE_MULTISEND as _DEFAULT_MULTISEND_ADDRESS,
)
from aea_test_autonomy.docker.registries import (
    SERVICE_MULTISIG_1 as _DEFAULT_SAFE_CONTRACT_ADDRESS,
)
from aea_test_autonomy.docker.registries import (
    SERVICE_REGISTRY as _DEFAULT_SERVICE_REGISTRY_ADDRESS,
)
from web3.types import Wei


TERMINATION_TIMEOUT = 120


class BaseTestTerminationEnd2End(
    BaseTestEnd2End
):  # pylint: disable=too-few-public-methods
    """
    Extended base class for conducting E2E tests with `termination_abci` activated.

    Test subclasses must set NB_AGENTS, agent_package, wait_to_finish and check_strings.
    """

    skill_package = "valory/register_termination_abci:0.1.0"
    SAFE_CONTRACT_ADDRESS = _DEFAULT_SAFE_CONTRACT_ADDRESS
    SERVICE_REGISTRY_ADDRESS = _DEFAULT_SERVICE_REGISTRY_ADDRESS
    MULTISEND_ADDRESS = _DEFAULT_MULTISEND_ADDRESS
    SERVICE_OWNER_ADDRESS, SERVICE_OWNER_PK = _HARDHAT_KEY_PAIRS[0]
    NETWORK_ENDPOINT = f"{_DEFAULT_HARDHAT_ADDR}:{_DEFAULT_HARDHAT_PORT}"
    CHAIN_ID = 31337
    SERVICE_ID = 1
    __args_prefix = f"vendor.valory.skills.{PublicId.from_str(skill_package).name}.models.params.args"
    extra_configs = [
        {
            "dotted_path": f"{__args_prefix}.service_registry_address",
            "value": SERVICE_REGISTRY_ADDRESS,
        },
        {
            "dotted_path": f"{__args_prefix}.on_chain_service_id",
            "value": SERVICE_ID,
        },
        {
            "dotted_path": f"{__args_prefix}.multisend_address",
            "value": MULTISEND_ADDRESS,
        },
    ]

    def _send_service_termination_signal(self) -> None:
        """
        Sends a termination signal to the service.

        The termination signal is just a zero transfer to the safe contract by the service.
        """
        instance = web3.Web3(web3.HTTPProvider(self.NETWORK_ENDPOINT))
        zero_eth = Wei(0)
        checksum_sender_address = instance.toChecksumAddress(self.SERVICE_OWNER_ADDRESS)
        checksum_receiver_address = instance.toChecksumAddress(
            self.SAFE_CONTRACT_ADDRESS
        )

        raw_tx = {
            "to": checksum_receiver_address,
            "from": checksum_sender_address,
            "value": zero_eth,
            "gas": 100000,
            "chainId": self.CHAIN_ID,
            "gasPrice": instance.eth.gas_price,
            "nonce": instance.eth.getTransactionCount(checksum_sender_address),
        }
        signed_tx = instance.eth.account.signTransaction(
            raw_tx, private_key=self.SERVICE_OWNER_PK
        )
        instance.eth.send_raw_transaction(signed_tx.rawTransaction)

    def test_run(self, nb_nodes: int) -> None:
        """Run the test."""
        self._send_service_termination_signal()
        self.prepare_and_launch(nb_nodes)
        self.health_check(
            max_retries=self.HEALTH_CHECK_MAX_RETRIES,
            sleep_interval=self.HEALTH_CHECK_SLEEP_INTERVAL,
        )
        self.check_aea_messages()
        self.terminate_agents(timeout=TERMINATION_TIMEOUT)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the round classes for register reset recovery."""

from enum import Enum
from typing import Dict, List, Optional, Set, Tuple, Type, cast

from packages.valory.skills.abstract_round_abci.base import (
    AbciApp,
    AbciAppTransitionFunction,
    AbstractRound,
    AppState,
    BaseSynchronizedData,
    CollectSameUntilThresholdRound,
)
from packages.valory.skills.register_reset_recovery_abci.payloads import (
    RoundCountPayload,
)


class Event(Enum):
    """Event enumeration for the Round Count round."""

    DONE = "done"
    ROUND_TIMEOUT = "round_timeout"


class RoundCountRound(CollectSameUntilThresholdRound):
    """A round in which the round count is stored as a list."""

    payload_class = RoundCountPayload
    synchronized_data_class = BaseSynchronizedData

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Event]]:
        """Process the end of the block."""
        if self.threshold_reached:
            # this simulates a state that is built across different rounds
            # we are using `round_count` here simply for convenience reasons,
            # it can be any data.
            all_round_counts: List[int] = cast(
                List[int], self.synchronized_data.db.get("round_counts", [])
            )
            all_round_counts.append(self.most_voted_payload)
            synchronized_data = self.synchronized_data.update(
                round_counts=all_round_counts,
            )
            return synchronized_data, Event.DONE
        return None


class RoundCountAbciApp(AbciApp[Event]):
    """RoundCountAbciApp

    Initial round: RoundCountRound

    Initial states: {RoundCountRound}

    Transition states:
        0. RoundCountRound
            - done: 0.

    Final states: {}

    Timeouts:
        round timeout: 30.0
    """

    initial_round_cls: Type[AbstractRound] = RoundCountRound
    initial_states = {RoundCountRound}
    transition_function: AbciAppTransitionFunction = {
        RoundCountRound: {
            Event.DONE: RoundCountRound,
        }
    }
    event_to_timeout: Dict[Event, float] = {
        Event.ROUND_TIMEOUT: 30.0,
    }
    db_pre_conditions: Dict[AppState, Set[str]] = {RoundCountRound: set()}



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

# pylint: disable=unused-argument,no-self-use

"""Tests for valory/test_ipfs_abci skill's behaviours."""

import json
import sys
from typing import Any, Callable, Generator, Optional
from unittest import mock
from unittest.mock import MagicMock

from aea_test_autonomy.helpers.base import try_send

from packages.valory.skills.abstract_round_abci.behaviour_utils import BaseBehaviour
from packages.valory.skills.abstract_round_abci.io_.store import SupportedObjectType
from packages.valory.skills.test_ipfs_abci.behaviours import DummyIpfsBehaviour


class TestDummyIpfsBehaviour:
    """Test for DummyIpfsBehaviour."""

    def wrap_dummy_get_from_ipfs(
        self, return_value: Optional[SupportedObjectType]
    ) -> Callable:
        """Wrap dummy_get_from_ipfs."""

        def dummy_get_from_ipfs(
            *args: Any, **kwargs: Any
        ) -> Generator[None, None, Optional[SupportedObjectType]]:
            """A mock get_from_ipfs."""
            return return_value
            yield

        return dummy_get_from_ipfs

    def wrap_dummy_send_to_ipfs(self, return_value: Optional[str]) -> Callable:
        """Wrap dummy_send_to_ipfs."""

        def dummy_send_to_ipfs(
            *args: Any, **kwargs: Any
        ) -> Generator[None, None, Optional[str]]:
            """A mock send_to_ipfs."""
            return return_value
            yield

        return dummy_send_to_ipfs

    def wrap_dummy_sleep(self) -> Callable:
        """Dummy sleep wrapper."""

        def dummy_sleep(*args: Any, **kwargs: Any) -> Generator:
            return
            yield

        return dummy_sleep

    def test_async_act(self) -> None:
        """Test DummyIpfsBehaviour.async_act"""
        behaviour = DummyIpfsBehaviour(name="", skill_context=MagicMock())
        gen = behaviour.async_act()
        dummy_object = {
            "dummy_k1": "dummy_v1",
            "dummy_k2": "dummy_v2",
            "dummy_k3": "dummy_v3",
        }
        with mock.patch.object(
            BaseBehaviour,
            "get_from_ipfs",
            side_effect=self.wrap_dummy_get_from_ipfs(dummy_object),
        ), mock.patch.object(
            BaseBehaviour,
            "send_to_ipfs",
            side_effect=self.wrap_dummy_send_to_ipfs(json.dumps(dummy_object)),
        ), mock.patch.object(
            BaseBehaviour, "sleep", side_effect=self.wrap_dummy_sleep()
        ), mock.patch.object(
            sys, "exit"
        ):
            try_send(gen)

    def test_async_act_multiple(self) -> None:
        """Test DummyIpfsBehaviour.async_act"""
        behaviour = DummyIpfsBehaviour(name="", skill_context=MagicMock())
        gen = behaviour.async_act()
        dummy_object = {
            "dummy_k1": "dummy_v1",
            "dummy_k2": "dummy_v2",
            "dummy_k3": "dummy_v3",
        }
        multiple_objects = {
            "test1.json": dummy_object,
            "test2.json": dummy_object,
            "test3.json": dummy_object,
        }
        with mock.patch.object(
            BaseBehaviour,
            "get_from_ipfs",
            side_effect=self.wrap_dummy_get_from_ipfs(multiple_objects),
        ), mock.patch.object(
            BaseBehaviour,
            "send_to_ipfs",
            side_effect=self.wrap_dummy_get_from_ipfs(multiple_objects),
        ), mock.patch.object(
            BaseBehaviour, "sleep", side_effect=self.wrap_dummy_sleep()
        ), mock.patch.object(
            sys, "exit"
        ):
            try_send(gen)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test `mint` command group."""

from typing import Any, Tuple
from unittest import mock

import pytest
from aea.configurations.data_types import PackageId, PackageType
from aea_test_autonomy.configurations import ETHEREUM_KEY_DEPLOYER
from aea_test_autonomy.fixture_helpers import registries_scope_class  # noqa: F401
from requests.exceptions import ConnectionError as RequestsConnectionError

from autonomy.chain.mint import registry_contracts

from tests.test_autonomy.test_chain.base import (
    BaseChainInteractionTest,
    DEFAULT_SERVICE_MINT_PARAMETERS,
    DUMMY_AGENT,
    DUMMY_CONNECTION,
    DUMMY_CONTRACT,
    DUMMY_PACKAGE_MANAGER,
    DUMMY_PROTOCOL,
    DUMMY_SERVICE,
    DUMMY_SKILL,
)


CUSTOM_OWNER = "0x8626F6940E2EB28930EFB4CEF49B2D1F2C9C1199"


class DummyContract:
    """Dummy contract"""

    def filter_token_id_from_emitted_events(self, *args: Any, **kwargs: Any) -> None:
        """Dummy method implementation"""

        raise RequestsConnectionError()

    def get_create_transaction(self, *args: Any, **kwargs: Any) -> None:
        """Dummy method implementation"""


class TestMintComponents(BaseChainInteractionTest):
    """Test `autonomy develop mint` command."""

    cli_options = ("mint",)

    @pytest.mark.parametrize(
        argnames=("package_id"),
        argvalues=(
            DUMMY_PROTOCOL,
            DUMMY_CONTRACT,
            DUMMY_CONNECTION,
            DUMMY_SKILL,
            DUMMY_AGENT,
        ),
    )
    def test_mint_components(self, package_id: PackageId) -> None:
        """Test mint components."""

        commands = [
            package_id.package_type.value,
            str(
                DUMMY_PACKAGE_MANAGER.package_path_from_package_id(
                    package_id=package_id
                )
            ),
            "--key",
            str(ETHEREUM_KEY_DEPLOYER),
        ]

        if package_id.package_type == PackageType.CONNECTION:
            # Dummy protocol is a dependency for the dummy connection
            commands += ["-d", str(self.mint_component(package_id=DUMMY_PROTOCOL))]

        if package_id.package_type == PackageType.SKILL:
            # Dummy protocol, dummy contract and dummy connection are dependencies for the dummy skill
            commands += [
                "-d",
                str(self.mint_component(package_id=DUMMY_PROTOCOL)),
                "-d",
                str(self.mint_component(package_id=DUMMY_CONTRACT)),
                "-d",
                str(self.mint_component(package_id=DUMMY_CONNECTION)),
            ]

        if package_id.package_type == PackageType.AGENT:
            # Dummy protocol, dummy contract, dummy connection and dummy skill are dependencies for the dummy skill
            commands += [
                "-d",
                str(self.mint_component(package_id=DUMMY_PROTOCOL)),
                "-d",
                str(self.mint_component(package_id=DUMMY_CONTRACT)),
                "-d",
                str(self.mint_component(package_id=DUMMY_CONNECTION)),
                "-d",
                str(self.mint_component(package_id=DUMMY_SKILL)),
            ]

        result = self.run_cli(commands=tuple(commands))

        assert result.exit_code == 0, result.stderr
        assert "Component minted with:" in result.output
        assert "Metadata Hash:" in result.output
        assert "Token ID:" in result.output

        token_id = self.extract_token_id_from_output(output=result.output)
        self.verify_minted_token_id(
            token_id=token_id,
            package_id=package_id,
        )
        self.verify_and_remove_metadata_file(token_id=token_id)

    def test_mint_component_with_owner(
        self,
    ) -> None:
        """Test mint components."""

        commands = [
            DUMMY_PROTOCOL.package_type.value,
            str(
                DUMMY_PACKAGE_MANAGER.package_path_from_package_id(
                    package_id=DUMMY_PROTOCOL
                )
            ),
            "--key",
            str(ETHEREUM_KEY_DEPLOYER),
            "--owner",
            CUSTOM_OWNER,
        ]

        result = self.run_cli(commands=tuple(commands))
        assert result.exit_code == 0, result.output
        assert "Component minted with:" in result.output
        assert "Metadata Hash:" in result.output
        assert "Token ID:" in result.output

        token_id = self.extract_token_id_from_output(output=result.output)
        self.verify_owner_address(
            token_id=token_id,
            owner=CUSTOM_OWNER,
            package_id=DUMMY_PROTOCOL,
        )
        self.verify_and_remove_metadata_file(token_id=token_id)

    def test_mint_service(
        self,
    ) -> None:
        """Test mint components."""
        with mock.patch("autonomy.cli.helpers.chain.verify_component_dependencies"):
            agent_id = self.mint_component(package_id=DUMMY_AGENT, dependencies=[1])

        commands = (
            DUMMY_SERVICE.package_type.value,
            str(
                DUMMY_PACKAGE_MANAGER.package_path_from_package_id(
                    package_id=DUMMY_SERVICE
                )
            ),
            "--key",
            str(ETHEREUM_KEY_DEPLOYER),
            "-a",
            str(agent_id),
            *DEFAULT_SERVICE_MINT_PARAMETERS[2:],
        )

        result = self.run_cli(commands=commands)

        assert result.exit_code == 0, result
        assert "Service minted with:" in result.output
        assert "Metadata Hash:" in result.output
        assert "Token ID:" in result.output

        token_id = self.extract_token_id_from_output(output=result.output)
        self.verify_minted_token_id(
            token_id=token_id,
            package_id=DUMMY_SERVICE,
        )
        self.verify_and_remove_metadata_file(token_id=token_id)

    def test_mint_service_with_owner(
        self,
    ) -> None:
        """Test mint components."""
        with mock.patch("autonomy.cli.helpers.chain.verify_component_dependencies"):
            agent_id = self.mint_component(package_id=DUMMY_AGENT, dependencies=[1])

        commands = (
            DUMMY_SERVICE.package_type.value,
            str(
                DUMMY_PACKAGE_MANAGER.package_path_from_package_id(
                    package_id=DUMMY_SERVICE
                )
            ),
            "--key",
            str(ETHEREUM_KEY_DEPLOYER),
            "-a",
            str(agent_id),
            *DEFAULT_SERVICE_MINT_PARAMETERS[2:],
            "--owner",
            CUSTOM_OWNER,
        )

        result = self.run_cli(commands=commands)

        assert result.exit_code == 0, result
        assert "Service minted with:" in result.output
        assert "Metadata Hash:" in result.output
        assert "Token ID:" in result.output

        token_id = self.extract_token_id_from_output(output=result.output)
        self.verify_owner_address(
            token_id=token_id,
            owner=CUSTOM_OWNER,
            package_id=DUMMY_SERVICE,
        )
        self.verify_and_remove_metadata_file(token_id=token_id)

    @pytest.mark.parametrize(
        argnames=("package_id", "parameters"),
        argvalues=(
            (DUMMY_CONNECTION, ("-d", "2")),
            (DUMMY_AGENT, ("-d", "1")),
            (DUMMY_SERVICE, ("-a", "1", "-n", "4", "-c", "1000", "--threshold", "3")),
        ),
    )
    def test_metadata_retrive_failure(
        self, package_id: PackageId, parameters: Tuple[str, ...]
    ) -> None:
        """Test connection error."""

        _ = registry_contracts.component_registry
        _ = registry_contracts.agent_registry
        _ = registry_contracts.service_registry

        with mock.patch.object(
            registry_contracts._component_registry,
            "get_token_uri",
            side_effect=RequestsConnectionError,
        ), mock.patch.object(
            registry_contracts._agent_registry,
            "get_token_uri",
            side_effect=RequestsConnectionError,
        ), mock.patch.object(
            registry_contracts._service_registry,
            "get_token_uri",
            side_effect=RequestsConnectionError,
        ):
            result = self.run_cli(
                commands=(
                    package_id.package_type.value,
                    str(
                        DUMMY_PACKAGE_MANAGER.package_path_from_package_id(
                            package_id=package_id
                        )
                    ),
                    "--key",
                    str(ETHEREUM_KEY_DEPLOYER),
                    *parameters,
                ),
            )
            self.cli_runner.mix_stderr = True
            assert result.exit_code == 1, result.output
            assert (
                "Dependency verification failed; Error connecting to the RPC"
                in result.stderr
            )

        with mock.patch.object(
            registry_contracts._component_registry, "get_token_uri"
        ), mock.patch.object(
            registry_contracts._agent_registry, "get_token_uri"
        ), mock.patch.object(
            registry_contracts._service_registry, "get_token_uri"
        ), mock.patch(
            "autonomy.chain.utils.r_get", side_effect=RequestsConnectionError
        ):
            result = self.run_cli(
                commands=(
                    package_id.package_type.value,
                    str(
                        DUMMY_PACKAGE_MANAGER.package_path_from_package_id(
                            package_id=package_id
                        )
                    ),
                    "--key",
                    str(ETHEREUM_KEY_DEPLOYER),
                    *parameters,
                ),
            )
            self.cli_runner.mix_stderr = True
            assert result.exit_code == 1, result.output
            assert (
                "Dependency verification failed; Error connecting to the IPFS gateway"
                in result.stdout
            )

    def test_connection_error(
        self,
    ) -> None:
        """Test connection error."""

        with mock.patch(
            "autonomy.chain.mint.transact", side_effect=RequestsConnectionError
        ):
            result = self.run_cli(
                commands=(
                    "protocol",
                    str(
                        DUMMY_PACKAGE_MANAGER.package_path_from_package_id(
                            package_id=DUMMY_PROTOCOL
                        )
                    ),
                    "--key",
                    str(ETHEREUM_KEY_DEPLOYER),
                ),
            )
            self.cli_runner.mix_stderr = True
            assert result.exit_code == 1, result.output
            assert (
                "Component mint failed with following error; Cannot connect to the given RPC"
                in result.stderr
            )

    def test_connection_error_service(
        self,
    ) -> None:
        """Test connection error."""

        with mock.patch(
            "autonomy.chain.mint.transact", side_effect=RequestsConnectionError
        ), mock.patch("autonomy.cli.helpers.chain.verify_service_dependencies"):
            result = self.run_cli(
                commands=(
                    "service",
                    str(
                        DUMMY_PACKAGE_MANAGER.package_path_from_package_id(
                            package_id=DUMMY_SERVICE
                        )
                    ),
                    "--key",
                    str(ETHEREUM_KEY_DEPLOYER),
                    *DEFAULT_SERVICE_MINT_PARAMETERS,
                ),
            )
            self.cli_runner.mix_stderr = True
            assert result.exit_code == 1, result.output
            assert (
                "Service mint failed with following error; Cannot connect to the given RPC"
                in result.stderr
            )

    def test_bad_owner_string(
        self,
    ) -> None:
        """Test connection error."""

        with mock.patch("autonomy.cli.helpers.chain.verify_service_dependencies"):
            result = self.run_cli(
                commands=(
                    "service",
                    str(
                        DUMMY_PACKAGE_MANAGER.package_path_from_package_id(
                            package_id=DUMMY_SERVICE
                        )
                    ),
                    "--key",
                    str(ETHEREUM_KEY_DEPLOYER),
                    *DEFAULT_SERVICE_MINT_PARAMETERS,
                    "--owner",
                    "0xowner",
                ),
            )
            self.cli_runner.mix_stderr = True
            assert result.exit_code == 1, result.output
            assert "Invalid owner address 0xowner" in result.stderr

    def test_fail_token_id_retrieve(
        self,
    ) -> None:
        """Test token id retrieval failure."""

        with mock.patch.object(
            registry_contracts, "_component_registry", DummyContract()
        ), mock.patch("autonomy.chain.mint.transact"):
            result = self.run_cli(
                commands=(
                    DUMMY_PROTOCOL.package_type.value,
                    str(
                        DUMMY_PACKAGE_MANAGER.package_path_from_package_id(
                            package_id=DUMMY_PROTOCOL
                        )
                    ),
                    "--key",
                    str(ETHEREUM_KEY_DEPLOYER),
                ),
            )

            assert result.exit_code == 1, result.output
            assert (
                "Component mint was successful but token ID retrieving failed with following error; "
                "Connection interrupted while waiting for the unitId emit event"
                in result.stderr
            )

    def test_fail_token_id_retrieve_service(
        self,
    ) -> None:
        """Test token id retrieval failure."""

        with mock.patch.object(
            registry_contracts, "_service_registry", DummyContract()
        ), mock.patch("autonomy.chain.mint.transact"), mock.patch(
            "autonomy.cli.helpers.chain.verify_service_dependencies"
        ):
            result = self.run_cli(
                commands=(
                    DUMMY_SERVICE.package_type.value,
                    str(
                        DUMMY_PACKAGE_MANAGER.package_path_from_package_id(
                            DUMMY_SERVICE
                        )
                    ),
                    "--key",
                    str(ETHEREUM_KEY_DEPLOYER),
                    *DEFAULT_SERVICE_MINT_PARAMETERS,
                ),
            )

            assert result.exit_code == 1, result.output
            assert (
                "Service mint was successful but token ID retrieving failed with following error; "
                "Connection interrupted while waiting for the unitId emit event"
                in result.stderr
            )

    def test_fail_dependency_does_not_match_service(
        self,
    ) -> None:
        """Test token id retrieval failure."""

        result = self.run_cli(
            commands=(
                DUMMY_SERVICE.package_type.value,
                str(DUMMY_PACKAGE_MANAGER.package_path_from_package_id(DUMMY_SERVICE)),
                "--key",
                str(ETHEREUM_KEY_DEPLOYER),
                *DEFAULT_SERVICE_MINT_PARAMETERS,
            ),
        )

        assert result.exit_code == 1, result.output
        assert (
            "On chain ID of the agent does not match with the one in the service configuration"
            in result.stderr
        )

    def test_fail_dependency_does_not_match_component(
        self,
    ) -> None:
        """Test token id retrieval failure."""

        result = self.run_cli(
            commands=(
                DUMMY_CONNECTION.package_type.value,
                str(
                    DUMMY_PACKAGE_MANAGER.package_path_from_package_id(DUMMY_CONNECTION)
                ),
                "--key",
                str(ETHEREUM_KEY_DEPLOYER),
                "-d",
                "1",
            ),
        )

        assert result.exit_code == 1, result.output
        assert (
            "On chain dependency with id 1 and public ID valory/abci:any not found in the local package configuration"
            in result.stderr
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Conftest module for Pytest."""
import logging
import socket
import time
import urllib
from typing import Generator, cast

import pytest
import requests

from deployments.Dockerfiles.tendermint.tendermint import (  # type: ignore
    DEFAULT_RPC_LISTEN_ADDRESS,
)


__parse_result = urllib.parse.urlparse(DEFAULT_RPC_LISTEN_ADDRESS)  # type: ignore
__IP, __RPC_PORT = __parse_result.hostname, __parse_result.port


@pytest.fixture(scope="session")
def http_() -> str:
    """The http prefix."""
    return "http://"


@pytest.fixture(scope="session")
def loopback() -> str:
    """The loopback address."""
    return "127.0.0.1"


@pytest.fixture(scope="session")
def rpc_port() -> int:
    """The Tendermint RPC port."""
    return cast(int, __RPC_PORT)


@pytest.fixture(scope="session")
def max_retries() -> int:
    """Max retries to make when ensuring Tendermint is running."""
    return 10


@pytest.fixture(scope="session")
def sleep_amount() -> int:
    """Amount to sleep in seconds."""
    return 1


def __port_is_open(ip: str, port: int) -> bool:
    """Assess whether a port is open"""
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    is_open = sock.connect_ex((ip, port)) == 0
    sock.close()
    return is_open


@pytest.fixture
def wait_for_node(
    http_: str, loopback: str, rpc_port: int, max_retries: int, sleep_amount: int
) -> Generator[None, None, None]:
    """Wait for Tendermint node to run."""
    i = 0
    while not __port_is_open(loopback, rpc_port) and i < max_retries:
        logging.debug(f"waiting for node... t={i}")
        i += 1
        time.sleep(sleep_amount)
    response = requests.get(f"{http_}{loopback}:{rpc_port}/status")
    success = response.status_code == 200
    assert success, "Tendermint node not running"
    yield


@pytest.fixture
def wait_for_occupied_rpc_port(
    http_: str, loopback: str, rpc_port: int, max_retries: int, sleep_amount: int
) -> Generator[None, None, None]:
    """Wait for Tendermint to occupy rpc_port."""
    i = 0
    while not __port_is_open(loopback, rpc_port):
        logging.debug(f"waiting for node... t={i}")
        i += 1
        time.sleep(sleep_amount)
        assert i < max_retries, "Failed to run tendermint node in time."
    yield



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the data classes for the Hello World ABCI application."""

from enum import Enum
from typing import Dict, Optional, Tuple

from packages.valory.skills.abstract_round_abci.base import (
    AbciApp,
    AbciAppTransitionFunction,
    AppState,
    BaseSynchronizedData,
    CollectDifferentUntilAllRound,
)
from packages.valory.skills.test_ipfs_abci.payloads import DummyPayload


class Event(Enum):
    """Event enumeration for the ipfs test skill."""

    DONE = "done"
    ROUND_TIMEOUT = "round_timeout"


class IpfsRound(CollectDifferentUntilAllRound):
    """Dummy ipfs round."""

    payload_class = DummyPayload
    synchronized_data_class = BaseSynchronizedData

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Enum]]:
        """End block. This is a dummy round, we don't care about payloads."""
        if self.collection_threshold_reached:  # pragma: no cover
            # added to satisfy fsm-spec checks,
            # the following will never be executed
            return self.synchronized_data, Event.DONE
        return None


class IpfsTestAbciApp(AbciApp[Event]):
    """IpfsTestAbciApp

    Initial round: IpfsRound

    Initial states: {IpfsRound}

    Transition states:
        0. IpfsRound
            - done: 0.
            - round timeout: 0.

    Final states: {}

    Timeouts:
        round timeout: 30.0
    """

    initial_round_cls: AppState = IpfsRound
    transition_function: AbciAppTransitionFunction = {
        IpfsRound: {
            Event.DONE: IpfsRound,
            Event.ROUND_TIMEOUT: IpfsRound,
        },
    }
    event_to_timeout: Dict[Event, float] = {
        Event.ROUND_TIMEOUT: 30.0,
    }



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tools for testing docker images"""

import contextlib
import json
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import docker
from docker.errors import APIError
from docker.models.containers import Container


class BaseImageBuildTest:
    """Test image build and run."""

    client: docker
    path: Path
    tag: str
    agent: str

    running_containers: List[Container]

    @classmethod
    def setup_class(cls) -> None:
        """Setup class."""

        cls.client = docker.from_env()

    def setup(self) -> None:
        """Setup test."""
        self.running_containers = []

    def build_image(
        self,
        path: Path,
        tag: str,
        buildargs: Optional[Dict[str, str]] = None,
    ) -> Tuple[bool, str]:
        """Build docker image."""

        stream = self.client.api.build(
            path=str(path),
            tag=tag,
            nocache=True,
            buildargs=(buildargs or {}),
        )
        output = ""
        for stream_obj in stream:
            for line in stream_obj.decode().split("\n"):
                line = line.strip()
                if not line:
                    continue
                stream_data = json.loads(line)
                if "stream" in stream_data:
                    output += stream_data["stream"]
                elif "errorDetail" in stream_data:
                    return False, stream_data["errorDetail"]["message"]
                elif "aux" in stream_data:
                    output += stream_data["aux"]["ID"]
                elif "status" in stream_data:
                    output += stream_data["status"]

        return True, output

    def teardown(self) -> None:
        """Teardown test."""

        for container in self.running_containers:
            with contextlib.suppress(APIError):
                container.kill()
                container.wait(timeout=30)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Helper for docstring analyser."""

import importlib
from pathlib import Path
from types import ModuleType
from typing import Optional, cast

import click
from aea.configurations.constants import PACKAGES

from autonomy.analyse.abci.docstrings import (
    ABCIAPP,
    compare_docstring_content,
    docstring_abci_app,
)


def import_rounds_module(
    module_path: Path,
    packages_dir: Optional[Path] = None,
) -> ModuleType:
    """Import module using importlib.import_module"""
    packages_dir = (packages_dir or Path.cwd() / PACKAGES).parent
    module_dir = module_path.parent.resolve().relative_to(packages_dir)
    import_path = ".".join((*module_dir.parts, "rounds"))
    return importlib.import_module(import_path)


def analyse_docstrings(
    module_path: Path,
    update: bool = False,
    packages_dir: Optional[Path] = None,
) -> bool:
    """
    Process module.

    :param module_path: Path to the rounds module
    :param update: whether to update the content if required
    :param packages_dir: Path to packages directory

    :return: boolean specifying whether the update is needed or not
    """

    module = import_rounds_module(module_path=module_path, packages_dir=packages_dir)
    for obj in dir(module):
        if obj.endswith(ABCIAPP) and obj != ABCIAPP:
            App = getattr(module, obj)
            docstring = docstring_abci_app(App)

            original_content = Path(cast(str, module.__file__)).read_text(
                encoding="utf-8"
            )
            has_docstring, expected_content = compare_docstring_content(
                file_content=original_content, docstring=docstring, abci_app_name=obj
            )

            if not has_docstring and update:
                click.echo(
                    f"App definition in {module_path} does not contain well formatted docstring, please update it manually"
                )
                return True

            update_needed = original_content != expected_content

            if update and update_needed:
                module_path.write_text(expected_content, encoding="utf-8")
                return True

            return update_needed

    click.echo(f"WARNING: No AbciApp definition found in: {module.__file__}")
    return True



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""The module contains test_helpers for module tests."""

from pathlib import Path
from types import ModuleType
from typing import cast
from unittest import mock

import pytest
from _pytest.capture import CaptureFixture  # type: ignore

from autonomy.cli.helpers import docstring
from autonomy.cli.helpers.docstring import analyse_docstrings, import_rounds_module

import packages
from packages.valory.skills import test_abci
from packages.valory.skills.test_abci import rounds as test_abci_rounds


@pytest.mark.parametrize("module", [test_abci, test_abci_rounds])
@pytest.mark.parametrize("with_package_dir", [True, False])
def test_import_rounds_module(module: ModuleType, with_package_dir: bool) -> None:
    """Test import_rounds_module"""

    packages_dir = (
        Path(cast(str, packages.__file__)).parent if with_package_dir else None
    )
    module_path = Path(cast(str, module.__file__))
    module = import_rounds_module(module_path, packages_dir)
    assert module is test_abci_rounds


def test_import_rounds_module_failure() -> None:
    """Test import_rounds_module"""

    packages_module = Path(cast(str, packages.__file__))
    module_path = Path(cast(str, test_abci.__file__))

    with pytest.raises(ModuleNotFoundError, match="No module named 'packages.rounds'"):
        import_rounds_module(packages_module)

    with pytest.raises(ModuleNotFoundError, match="No module named 'skills'"):
        import_rounds_module(module_path, packages_dir=module_path.parent.parent)


@pytest.mark.parametrize("module", [test_abci, test_abci_rounds])
def test_analyse_docstrings_without_update(module: ModuleType) -> None:
    """Test analyse_docstrings"""

    module_path = Path(cast(str, module.__file__))
    updated_needed = analyse_docstrings(module_path)
    assert not updated_needed


def test_analyse_docstrings_no_abci_app_definition(capsys: CaptureFixture) -> None:
    """Test analyse_docstrings no ABCIApp definition found"""

    with mock.patch.object(docstring, "import_rounds_module", return_value=docstring):
        module_path = Path(cast(str, test_abci.__file__))
        updated_needed = analyse_docstrings(module_path)
        stdout = capsys.readouterr().out
        expected = f"WARNING: No AbciApp definition found in: {docstring.__file__}"
        assert updated_needed
        assert expected in stdout


def test_analyse_docstrings_with_update(capsys: CaptureFixture) -> None:
    """Test analyse_docstrings with update"""

    module_path = Path(cast(str, test_abci_rounds.__file__))
    doc = cast(str, test_abci_rounds.TestAbciApp.__doc__)
    content_with_mutated_abci_doc = module_path.read_text().replace(doc, doc + " ")

    with mock.patch.object(Path, "write_text") as mock_write_text:
        with mock.patch.object(Path, "read_text", return_value=""):
            updated_needed = analyse_docstrings(module_path, update=True)
            assert updated_needed
            out, _ = capsys.readouterr()
            expected = (
                "does not contain well formatted docstring, please update it manually"
            )
            assert expected in out

        with mock.patch.object(
            Path, "read_text", return_value=content_with_mutated_abci_doc
        ):
            updated_needed = analyse_docstrings(module_path, update=True)
            assert updated_needed
            mock_write_text.assert_called_once()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------


"""Tests for the `autonomy analyse logs` command"""

import contextlib
import os
from typing import Tuple

from autonomy.analyse.logs.base import LOGS_DB
from autonomy.deploy.constants import WARNING

from tests.conftest import DATA_DIR
from tests.test_autonomy.test_cli.base import BaseCliTest


LOGS_DIR = DATA_DIR / "logs"
LOGS_DB_FILE = LOGS_DIR / LOGS_DB

AVAILABLE_ROUNDS = (
    "registration_startup",
    "new_tokens",
    "leaderboard_observation",
    "image_code_calculation",
    "image_generation",
    "db_update",
    "reset_and_pause",
)


class BaseLogAnalyserTest(BaseCliTest):
    """Base test class for the log analyser."""

    cli_options: Tuple[str, ...] = ("analyse", "logs")

    @classmethod
    def teardown_class(cls) -> None:
        """Teardown class."""

        with contextlib.suppress(PermissionError):
            if LOGS_DB_FILE.exists():
                os.remove(LOGS_DB_FILE)


class TestAnalyseLogs(BaseLogAnalyserTest):
    """Test `autonomy analyse logs`"""

    def test_agent_options(self) -> None:
        """Test print agent options."""

        result = self.run_cli(commands=("--from-dir", str(LOGS_DIR)))

        assert result.exit_code == 1, result.stdout
        assert "Available agents: ['aea_0']" in result.output

    def test_logs_table(self) -> None:
        """Test print agent options."""

        result = self.run_cli(commands=("--from-dir", str(LOGS_DIR), "-a", "aea_0"))

        assert result.exit_code == 0, result.stdout
        assert "Entered in the 'registration_startup' behaviour" in result.output

    def test_log_level_filter(self) -> None:
        """Test print agent options."""

        result = self.run_cli(
            commands=(
                "--from-dir",
                str(LOGS_DIR),
                "-a",
                "aea_0",
                "--log-level",
                WARNING,
            )
        )

        assert result.exit_code == 0, result.stdout
        assert "[WARNING]" in result.output
        assert "[INFO]" not in result.output

    def test_start_time_filter(self) -> None:
        """Test print agent options."""

        result = self.run_cli(
            commands=(
                "--from-dir",
                str(LOGS_DIR),
                "-a",
                "aea_0",
                "--start-time",
                "2023-01-23 15:39:20,424",  # actual timestamp from the logs file
            )
        )

        assert result.exit_code == 0, result.stdout

        # actual timestamps from the logs file
        assert "2023-01-23 15:39:21.441000" in result.output
        assert "2023-01-23 15:39:19.686000" not in result.output

    def test_end_time_filter(self) -> None:
        """Test print agent options."""

        result = self.run_cli(
            commands=(
                "--from-dir",
                str(LOGS_DIR),
                "-a",
                "aea_0",
                "--end-time",
                "2023-01-23 15:39:20,424",  # actual timestamp from the logs file
            )
        )

        assert result.exit_code == 0, result.stdout

        # actual timestamps from the logs file
        assert "2023-01-23 15:39:21.441000" not in result.output
        assert "2023-01-23 15:39:19.686000" in result.output

    def test_period_filter(self) -> None:
        """Test print agent options."""

        result = self.run_cli(
            commands=(
                "--from-dir",
                str(LOGS_DIR),
                "-a",
                "aea_0",
                "--period",
                "1",
            )
        )

        assert result.exit_code == 0, result.stdout

        assert "period 0" not in result.output
        assert "period 1" in result.output
        assert "period 2" not in result.output

    def test_round_filter(self) -> None:
        """Test print agent options."""

        round_name = "new_tokens"
        result = self.run_cli(
            commands=(
                "--from-dir",
                str(LOGS_DIR),
                "-a",
                "aea_0",
                "--round",
                round_name,
            )
        )
        assert result.exit_code == 0, result.stdout

        assert round_name in result.output
        assert all([_round not in result.output for _round in AVAILABLE_ROUNDS[2:]])

    def test_include_regex_filter(self) -> None:
        """Test print agent options."""

        result = self.run_cli(
            commands=(
                "--from-dir",
                str(LOGS_DIR),
                "-a",
                "aea_0",
                "-ir",
                ".*Entered in the.*",
            )
        )
        assert result.exit_code == 0, result.stdout
        assert "Entered in the" in result.output
        assert "arrived block with timestamp" not in result.output

    def test_exclude_regex_filter(self) -> None:
        """Test print agent options."""

        result = self.run_cli(
            commands=(
                "--from-dir",
                str(LOGS_DIR),
                "-a",
                "aea_0",
                "-er",
                ".*Entered in the.*",
            )
        )
        assert result.exit_code == 0, result.stdout
        assert "Entered in the" not in result.output
        assert "arrived block with timestamp" in result.output

    def test_fsm_path(self) -> None:
        """Test print agent options."""

        result = self.run_cli(
            commands=(
                "--from-dir",
                str(LOGS_DIR),
                "-a",
                "aea_0",
                "--fsm",
            )
        )
        assert result.exit_code == 0, result.stdout
        assert all(
            [
                f"|_ {_round} | Event.DONE" in result.output
                for _round in AVAILABLE_ROUNDS
            ]
        )

    def test_empty_logs_dir(self) -> None:
        """Test print agent options."""

        result = self.run_cli(
            commands=(
                "--from-dir",
                str(LOGS_DIR.parent),
                "-a",
                "aea_0",
                "--fsm",
            )
        )
        assert result.exit_code == 1, result.stdout
        assert "Cannot find agent log data in" in result.output



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Utility functions."""
from json import JSONDecodeError
from typing import Dict, List

from aea.configurations.base import PackageConfiguration
from aea.configurations.data_types import PackageId, PublicId
from aea.crypto.base import LedgerApi
from requests import get as r_get
from requests.exceptions import ConnectionError as RequestConnectionError

from autonomy.chain.base import registry_contracts
from autonomy.chain.exceptions import DependencyError, FailedToRetrieveComponentMetadata
from autonomy.chain.metadata import IPFS_URI_PREFIX
from autonomy.configurations.base import Service


def get_ipfs_hash_from_uri(uri: str) -> str:
    """Split IPFS hash from the ipfs uri"""

    return uri.replace(IPFS_URI_PREFIX, "")


def resolve_component_id(
    ledger_api: LedgerApi,
    contract_address: str,
    token_id: int,
    is_agent: bool = False,
    is_service: bool = False,
) -> Dict:
    """Resolve component ID to metadata json"""

    try:
        if is_service:
            token_uri_callable = registry_contracts.service_registry.get_token_uri
        elif is_agent:
            token_uri_callable = registry_contracts.agent_registry.get_token_uri
        else:
            token_uri_callable = registry_contracts.component_registry.get_token_uri

        metadata_uri = token_uri_callable(
            ledger_api=ledger_api,
            contract_address=contract_address,
            token_id=token_id,
        )
    except RequestConnectionError as e:
        raise FailedToRetrieveComponentMetadata("Error connecting to the RPC") from e

    try:
        return r_get(url=metadata_uri).json()
    except RequestConnectionError as e:
        raise FailedToRetrieveComponentMetadata(
            "Error connecting to the IPFS gateway"
        ) from e
    except JSONDecodeError as e:
        raise FailedToRetrieveComponentMetadata(
            f"Error decoding json data; make sure metadata file for the component exist on the IPFS registry; Dependency ID: {token_id}"
        ) from e


def parse_public_id_from_metadata(id_string: str) -> PublicId:
    """Parse public ID from on-chain metadata."""

    if ":" in id_string:
        id_string, _ = id_string.split(":")

    try:
        # author/package_name
        return PublicId.from_str(id_string).to_any()
    except ValueError as e:
        id_parts = id_string.split("/")

        if len(id_parts) == 3:
            # component_type/author/name
            _, author, name = id_parts
            return PublicId(author=author, name=name).to_any()

        if len(id_parts) == 4:
            # component_type/author/name/version
            return PackageId.from_uri_path(id_string).public_id.to_any()

        raise DependencyError(f"Invalid package name found `{id_string}`") from e


def verify_component_dependencies(
    ledger_api: LedgerApi,
    contract_address: str,
    dependencies: List[int],
    package_configuration: PackageConfiguration,
    skip_hash_check: bool = False,
) -> None:
    """Verify package dependencies using on-chain metadata."""

    public_id_to_hash: Dict[PublicId, List[str]] = {}

    for dependency in package_configuration.package_dependencies:
        public_id = dependency.public_id.to_any()
        if public_id not in public_id_to_hash:
            public_id_to_hash[public_id] = []
        public_id_to_hash[dependency.public_id.to_any()].append(dependency.package_hash)

    for dependency_id in dependencies:
        component_metadata = resolve_component_id(
            contract_address=contract_address,
            ledger_api=ledger_api,
            token_id=dependency_id,
        )
        component_public_id = parse_public_id_from_metadata(component_metadata["name"])
        if component_public_id not in public_id_to_hash:
            raise DependencyError(
                f"On chain dependency with id {dependency_id} and public ID {component_public_id} not found in the local package configuration"
            )

        if skip_hash_check and len(public_id_to_hash[component_public_id]) > 0:
            public_id_to_hash[component_public_id].pop()
        else:
            on_chain_hash = get_ipfs_hash_from_uri(uri=component_metadata["code_uri"])
            if on_chain_hash not in public_id_to_hash[component_public_id]:
                raise DependencyError(
                    f"Package hash does not match for the on chain package and the local package; Dependency={dependency_id}"
                )
            public_id_to_hash[component_public_id] = [
                _hash
                for _hash in public_id_to_hash[component_public_id]
                if _hash != on_chain_hash
            ]

        if len(public_id_to_hash[component_public_id]) == 0:
            del public_id_to_hash[component_public_id]

    if len(public_id_to_hash):
        missing_deps = list(map(str, public_id_to_hash.keys()))
        raise DependencyError(
            f"Please provide on chain ID as dependency for following packages; {missing_deps}"
        )


def verify_service_dependencies(
    ledger_api: LedgerApi,
    contract_address: str,
    agent_id: int,
    service_configuration: Service,
    skip_hash_check: bool = False,
) -> None:
    """Verify package dependencies using on-chain metadata."""

    agent = service_configuration.agent
    component_metadata = resolve_component_id(
        contract_address=contract_address,
        ledger_api=ledger_api,
        token_id=agent_id,
        is_agent=True,
    )
    component_public_id = parse_public_id_from_metadata(component_metadata["name"])
    if component_public_id != agent.to_any():
        raise DependencyError(
            "On chain ID of the agent does not match with the one in the service configuration"
        )

    if skip_hash_check:
        return

    if agent.hash != get_ipfs_hash_from_uri(uri=component_metadata["code_uri"]):
        raise DependencyError(
            f"Package hash does not match for the on chain package and the local package; Dependency={agent}"
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Scaffold skill from an FSM"""


import itertools
import os
import re
import shutil
from pathlib import Path
from typing import Dict, List, Optional, Type

import click
from aea.cli.fingerprint import fingerprint_item
from aea.cli.utils.context import Context
from aea.configurations.base import (
    AgentConfig,
    SkillComponentConfiguration,
    SkillConfig,
)
from aea.configurations.constants import (
    DEFAULT_AEA_CONFIG_FILE,
    DEFAULT_SKILL_CONFIG_FILE,
    SKILL,
    SKILLS,
)

# the decoration does side-effect on the 'aea scaffold' command
from aea.configurations.data_types import CRUDCollection, PackageType, PublicId
from aea.helpers.ipfs.base import IPFSHashOnly
from aea.package_manager.v1 import PackageManagerV1

from autonomy.analyse.abci.app_spec import DFA, FSMSpecificationLoader
from autonomy.configurations.constants import INIT_PY, PYCACHE
from autonomy.constants import ABSTRACT_ROUND_ABCI_SKILL_WITH_HASH
from autonomy.fsm.scaffold.base import AbstractFileGenerator
from autonomy.fsm.scaffold.constants import ABCI_APP, ROUND_BEHAVIOUR
from autonomy.fsm.scaffold.generators.components import (
    BehaviourFileGenerator,
    DialoguesFileGenerator,
    HandlersFileGenerator,
    ModelsFileGenerator,
    PayloadsFileGenerator,
    RoundFileGenerator,
)
from autonomy.fsm.scaffold.generators.tests import (
    BehaviourTestsFileGenerator,
    DialoguesTestFileGenerator,
    HandlersTestFileGenerator,
    ModelTestFileGenerator,
    PayloadTestsFileGenerator,
    RoundTestsFileGenerator,
)
from autonomy.fsm.scaffold.templates import COPYRIGHT_HEADER


TO_LOCAL_REGISTRY_FLAG = "to_local_registry"

ABSTRACT_ROUND_SKILL_PUBLIC_ID = PublicId.from_str(ABSTRACT_ROUND_ABCI_SKILL_WITH_HASH)


class SkillConfigUpdater:  # pylint: disable=too-few-public-methods
    """Update the skill configuration according to the Abci classes."""

    def __init__(self, ctx: Context, skill_dir: Path, dfa: DFA) -> None:
        """
        Initialize the skill config updater.

        :param ctx: the AEA CLI context object.
        :param skill_dir: the directory of the AEA skill package.
        :param dfa: the DFA object.
        """
        self.ctx = ctx
        self.skill_dir = skill_dir
        self.dfa = dfa

        self.skill_config_path = self.skill_dir / DEFAULT_SKILL_CONFIG_FILE

    def update(self) -> None:
        """Update the skill configuration file."""
        config = self.ctx.skill_loader.load(self.skill_config_path.open())
        self._update_behaviours(config)
        self._update_handlers(config)
        self._update_models(config)
        self._update_dependencies(config)
        self.ctx.skill_loader.dump(config, self.skill_config_path.open("w"))
        # TODO: update fingerprint_item to use path instead of context
        preserve_cwd = self.ctx.cwd
        if self.ctx.config.get(TO_LOCAL_REGISTRY_FLAG):
            self.ctx.cwd = Path(self.ctx.registry_path) / self.ctx.agent_config.author
        fingerprint_item(self.ctx, SKILL, config.public_id)
        self.ctx.cwd = preserve_cwd

    def _update_behaviours(self, config: SkillConfig) -> None:
        """Update the behaviours section of the skill configuration."""
        config.behaviours = CRUDCollection[SkillComponentConfiguration]()
        abci_app_cls_name = self.dfa.label.split(".")[-1]
        round_behaviour_cls_name = abci_app_cls_name.replace(ABCI_APP, ROUND_BEHAVIOUR)
        main_config = SkillComponentConfiguration(round_behaviour_cls_name)
        config.behaviours.create("main", main_config)

    def _update_handlers(  # pylint: disable=no-self-use
        self, config: SkillConfig
    ) -> None:
        """Update the handlers section of the skill configuration."""
        config.handlers = CRUDCollection[SkillComponentConfiguration]()
        config.handlers.create("abci", SkillComponentConfiguration("ABCIHandler"))
        config.handlers.create(
            "contract_api", SkillComponentConfiguration("ContractApiHandler")
        )
        config.handlers.create("http", SkillComponentConfiguration("HttpHandler"))
        config.handlers.create(
            "ledger_api", SkillComponentConfiguration("LedgerApiHandler")
        )
        config.handlers.create("signing", SkillComponentConfiguration("SigningHandler"))
        config.handlers.create(
            "tendermint", SkillComponentConfiguration("TendermintHandler")
        )
        config.handlers.create("ipfs", SkillComponentConfiguration("IpfsHandler"))

    def _update_models(  # pylint: disable=no-self-use
        self, config: SkillConfig
    ) -> None:
        """Update the models section of the skill configuration."""
        config.models = CRUDCollection[SkillComponentConfiguration]()
        config.models.create("state", SkillComponentConfiguration("SharedState"))
        config.models.create("requests", SkillComponentConfiguration("Requests"))
        config.models.create(
            "params",
            SkillComponentConfiguration("Params", **self._default_params_config),
        )
        config.models.create(
            "abci_dialogues", SkillComponentConfiguration("AbciDialogues")
        )
        config.models.create(
            "benchmark_tool",
            SkillComponentConfiguration("BenchmarkTool", log_dir="/logs"),
        )
        config.models.create(
            "http_dialogues", SkillComponentConfiguration("HttpDialogues")
        )
        config.models.create(
            "signing_dialogues", SkillComponentConfiguration("SigningDialogues")
        )
        config.models.create(
            "ledger_api_dialogues", SkillComponentConfiguration("LedgerApiDialogues")
        )
        config.models.create(
            "contract_api_dialogues",
            SkillComponentConfiguration("ContractApiDialogues"),
        )
        config.models.create(
            "tendermint_dialogues", SkillComponentConfiguration("TendermintDialogues")
        )
        config.models.create(
            "ipfs_dialogues", SkillComponentConfiguration("IpfsDialogues")
        )

    @classmethod
    def get_actual_abstract_round_abci_package_public_id(
        cls, ctx: Context
    ) -> Optional[PublicId]:
        """Get abstract round abci pacakge id from the registry."""
        package_manager = PackageManagerV1.from_dir(Path(ctx.registry_path))
        packages = [
            package_id.public_id
            for package_id in itertools.chain(
                package_manager.third_party_packages.keys(),
                package_manager.dev_packages.keys(),
            )
            if (
                package_id.public_id.without_hash().to_any()
                == ABSTRACT_ROUND_SKILL_PUBLIC_ID.without_hash().to_any()
                and package_id.package_type == PackageType.SKILL
            )
        ]
        if not packages:
            return None
        abstract_round_abci, *_ = packages
        return abstract_round_abci

    def _update_dependencies(self, config: SkillConfig) -> None:
        """Update skill dependencies."""
        # retrieve the actual valory/abstract_round_abci package

        if self.ctx.config.get(TO_LOCAL_REGISTRY_FLAG):
            abstract_round_abci = self.get_actual_abstract_round_abci_package_public_id(
                self.ctx
            )
            if not abstract_round_abci:
                raise ValueError("valory/abstract_round_abci package not found")
        else:
            agent_config = self._load_agent_config()
            abstract_round_abci = [
                public_id
                for public_id in agent_config.skills
                if public_id.author == ABSTRACT_ROUND_SKILL_PUBLIC_ID.author
                and public_id.name == ABSTRACT_ROUND_SKILL_PUBLIC_ID.name
            ][0]

        config.skills.add(abstract_round_abci)

    def _load_agent_config(self) -> AgentConfig:
        """Load the current agent configuration."""
        agent_config_path = Path(self.ctx.cwd) / DEFAULT_AEA_CONFIG_FILE
        with agent_config_path.open() as f:
            return self.ctx.agent_loader.load(f)

    @property
    def _default_params_config(self) -> Dict:
        """The default `params` configuration."""
        abci_app_cls_name = self.dfa.label.split(".")[-1]
        service_id = abci_app_cls_name.replace(ABCI_APP, "")
        service_id = re.sub(r"(?<!^)(?=[A-Z])", "_", service_id).lower()
        return {
            "cleanup_history_depth": 1,
            "cleanup_history_depth_current": None,
            "drand_public_key": "868f005eb8e6e4ca0a47c8a77ceaa5309a47978a7c71bc5cce96366b5d7a569937c529eeda66c7293784a9402801af31",
            "finalize_timeout": 60.0,
            "genesis_config": {
                "genesis_time": "2022-05-20T16:00:21.735122717Z",
                "chain_id": "chain-c4daS1",
                "consensus_params": {
                    "block": {
                        "max_bytes": "22020096",
                        "max_gas": "-1",
                        "time_iota_ms": "1000",
                    },
                    "evidence": {
                        "max_age_num_blocks": "100000",
                        "max_age_duration": "172800000000000",
                        "max_bytes": "1048576",
                    },
                    "validator": {"pub_key_types": ["ed25519"]},
                    "version": {},
                },
                "voting_power": "10",
            },
            "history_check_timeout": 1205,
            "ipfs_domain_name": None,
            "keeper_allowed_retries": 3,
            "keeper_timeout": 30.0,
            "max_attempts": 10,
            "max_healthcheck": 120,
            "reset_pause_duration": 10,
            "on_chain_service_id": None,
            "request_retry_delay": 1.0,
            "request_timeout": 10.0,
            "reset_tendermint_after": 2,
            "retry_attempts": 400,
            "retry_timeout": 3,
            "round_timeout_seconds": 30.0,
            "service_id": service_id,
            "service_registry_address": None,
            "setup": {
                "all_participants": ["0x0000000000000000000000000000000000000000"],
                "safe_contract_address": "0x0000000000000000000000000000000000000000",
                "consensus_threshold": None,
            },
            "share_tm_config_on_startup": False,
            "sleep_time": 1,
            "tendermint_check_sleep_delay": 3,
            "tendermint_com_url": "http://localhost:8080",
            "tendermint_max_retries": 5,
            "tendermint_p2p_url": "localhost:26656",
            "tendermint_url": "http://localhost:26657",
            "tx_timeout": 10.0,
            "validate_timeout": 1205,
        }


class ScaffoldABCISkill:
    """Utility class that implements the scaffolding of the ABCI skill."""

    file_generators: List[Type[AbstractFileGenerator]] = [
        RoundFileGenerator,
        BehaviourFileGenerator,
        PayloadsFileGenerator,
        ModelsFileGenerator,
        HandlersFileGenerator,
        DialoguesFileGenerator,
    ]

    test_file_generators: List[Type[AbstractFileGenerator]] = [
        RoundTestsFileGenerator,
        BehaviourTestsFileGenerator,
        PayloadTestsFileGenerator,
        ModelTestFileGenerator,
        HandlersTestFileGenerator,
        DialoguesTestFileGenerator,
    ]

    def __init__(self, ctx: Context, skill_name: str, spec_path: Path) -> None:
        """Initialize the utility class."""
        self.ctx = ctx
        self.skill_name = skill_name

        self.spec_path = spec_path

        # process FSM specification
        self.dfa = DFA.load(
            file=spec_path, spec_format=FSMSpecificationLoader.OutputFormats.YAML
        )

    @property
    def skill_dir(self) -> Path:
        """Get the directory to the skill."""
        if self.ctx.config.get(TO_LOCAL_REGISTRY_FLAG):
            return Path(
                self.ctx.registry_path,
                self.ctx.agent_config.author,
                SKILLS,
                self.skill_name,
            )
        return Path(self.ctx.cwd, SKILLS, self.skill_name)

    @property
    def skill_test_dir(self) -> Path:
        """Get the directory to the skill tests."""
        return self.skill_dir / "tests"

    def do_scaffolding(self) -> None:
        """Do the scaffolding."""
        self.skill_dir.mkdir(parents=True, exist_ok=True)
        self.skill_test_dir.mkdir(parents=True, exist_ok=True)

        file_dirs = self.skill_dir, self.skill_test_dir
        file_gens = self.file_generators, self.test_file_generators
        for (f_dir, gens) in zip(file_dirs, file_gens):
            for f_gen in gens:
                click.echo(f"Generating module {f_gen.FILENAME}...")
                f_gen(self.ctx, self.skill_name, self.dfa).write_file(f_dir)

        # remove original 'my_model.py' file
        os.remove(self.skill_dir / "my_model.py")

        self._remove_pycache()
        self._update_init_py()
        self._copy_spec_file()
        self._update_config()

    def add_skill_to_packages(self) -> None:
        """Add skill to packages.json if scaffolded to local packages repository"""

        click.echo("Adding skill package to `packages.json` file...")
        config = self.ctx.skill_loader.load(
            (self.skill_dir / DEFAULT_SKILL_CONFIG_FILE).open()
        )
        package_manager = PackageManagerV1.from_dir(Path(self.ctx.registry_path))
        package_manager.dev_packages[config.package_id] = IPFSHashOnly.hash_directory(
            dir_path=self.skill_dir,
        )
        package_manager.dump()

    def _update_config(self) -> None:
        """Update the skill configuration."""
        click.echo("Updating skill configuration...")
        SkillConfigUpdater(self.ctx, self.skill_dir, self.dfa).update()

    def _copy_spec_file(self) -> None:
        """Copy the spec file in the skill directory."""
        click.echo("Copying the spec file in the skill directory...")
        shutil.copy(self.spec_path, self.skill_dir / self.spec_path.name)

    def _update_init_py(self) -> None:
        """Update Copyright __init__.py files"""

        init_py_path = self.skill_dir / INIT_PY
        lines = init_py_path.read_text().splitlines()
        content = "\n".join(line for line in lines if not line.startswith("#"))
        init_py_path.write_text(f"{COPYRIGHT_HEADER} {content}\n")
        (Path(self.skill_test_dir) / INIT_PY).write_text(COPYRIGHT_HEADER)

    def _remove_pycache(self) -> None:
        """Remove __pycache__ folders."""
        for path in self.skill_dir.rglob(f"*{PYCACHE}*"):
            shutil.rmtree(path, ignore_errors=True)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
# pylint: disable=unused-import

"""Test the models.py module of the skill."""


from packages.valory.skills.abstract_round_abci.test_tools.base import DummyContext
from packages.valory.skills.termination_abci.models import SharedState


class TestSharedState:  # pylint: disable=too-few-public-methods
    """Test SharedState(Model) class."""

    def test_initialization(  # pylint: disable=no-self-use
        self,
    ) -> None:
        """Test initialization."""
        SharedState(name="", skill_context=DummyContext())
